# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for contacts."""

from absl.testing import absltest
from brax.v2 import geometry
from brax.v2 import kinematics
from brax.v2 import test_utils
from brax.v2.io import mjcf
from jax import numpy as jp
import numpy as np


class SphereTest(absltest.TestCase):

  def test_sphere_plane(self):
    sys = test_utils.load_fixture('ant.xml')
    # all 4 feet are just barely contacting the floor
    q = jp.array([0, 0, 0.556008, 1, 0, 0, 0, 0, 1, 0, 1, 0, 1, 0, 1])
    x, _ = kinematics.forward(sys, q, jp.zeros(sys.qd_size()))
    c = geometry.contact(sys, x)
    self.assertEqual(c.pos.shape[0], 4)  # 4 contacts

    # front right foot is in +x, +y quadrant and is just resting on the surface
    c = c.take(0)
    np.testing.assert_array_almost_equal(c.pos, jp.array([0.61612, 0.61612, 0]))
    np.testing.assert_array_almost_equal(c.normal, jp.array([0, 0, 1]))
    np.testing.assert_array_almost_equal(c.penetration, 0)
    np.testing.assert_array_almost_equal(c.friction, 1)
    np.testing.assert_array_almost_equal(c.elasticity, 0)
    self.assertEqual(c.link_idx, (2, -1))

  _SPHERE_SPHERE = """
    <mujoco model="two_spheres">
      <worldbody>
        <body name="body1" pos="0 0 0">
          <joint axis="1 0 0" name="free1" pos="0 0 0" type="free"/>
          <geom name="sphere1" pos="0 0 0" size="0.2" type="sphere"/>
        </body>
        <body name="body2" pos="0 0 0">
          <joint axis="1 0 0" name="free2" pos="0 0 0" type="free"/>
          <geom name="sphere2" pos="0 0.3 0" size="0.11" type="sphere"/>
        </body>
      </worldbody>
      <contact>
        <pair geom1="sphere1" geom2="sphere2"/>
      </contact>
    </mujoco>
  """

  def test_sphere_sphere(self):
    sys = mjcf.loads(self._SPHERE_SPHERE)
    x, _ = kinematics.forward(sys, sys.init_q, jp.zeros(sys.qd_size()))
    c = geometry.contact(sys, x).take(0)

    np.testing.assert_array_almost_equal(c.penetration, 0.01)
    np.testing.assert_array_almost_equal(c.pos, jp.array([0.0, 0.195, 0.0]))
    np.testing.assert_array_almost_equal(c.normal, jp.array([0, -1.0, 0.0]))

  _SPHERE_CAP = """
    <mujoco model="sphere_cap">
      <worldbody>
        <body name="body1" pos="0 0 0">
          <joint axis="1 0 0" name="free1" pos="0 0 0" type="free"/>
          <geom name="sphere1" pos="0 0.3 0" size="0.05" type="sphere"/>
        </body>
        <body name="body2" pos="0 0 0">
          <joint axis="1 0 0" name="free2" pos="0 0 0" type="free"/>
          <geom name="cap2" fromto="0.0 -0.5 0.14 0.0 0.5 0.14" pos="0 0 0" size="0.1" type="capsule"/>
        </body>
      </worldbody>
      <contact>
        <pair geom1="sphere1" geom2="cap2"/>
      </contact>
    </mujoco>
  """

  def test_sphere_capsule(self):
    sys = mjcf.loads(self._SPHERE_CAP)
    x, _ = kinematics.forward(sys, sys.init_q, jp.zeros(sys.qd_size()))
    c = geometry.contact(sys, x).take(0)

    np.testing.assert_array_almost_equal(c.penetration, 0.01)
    np.testing.assert_array_almost_equal(c.pos, jp.array([0.0, 0.3, 0.045]))
    np.testing.assert_array_almost_equal(c.normal, jp.array([0, 0.0, -1.0]))

  _SPHERE_BOX = """
    <mujoco model="sphere_box">
      <worldbody>
        <body name="body1" pos="0 0 0.54">
          <joint axis="1 0 0" name="free1" pos="0 0 0" type="free"/>
          <geom name="sphere1" pos="0 0 0" size="0.05" type="sphere"/>
        </body>
        <body name="body2" pos="0 0 0" euler="0 0 0">
          <joint axis="1 0 0" name="free2" pos="0 0 0" type="free"/>
          <geom name="box" pos="0 0 0" size="0.5 0.5 0.5" type="box"/>
        </body>
      </worldbody>
    </mujoco>
  """

  def test_sphere_mesh(self):
    sys = mjcf.loads(self._SPHERE_BOX)
    x, _ = kinematics.forward(sys, sys.init_q, jp.zeros(sys.qd_size()))
    c = geometry.contact(sys, x)

    np.testing.assert_array_almost_equal((c.penetration > 0).sum(), 2)
    c = c.take(c.penetration.argsort()[-2:])
    np.testing.assert_array_almost_equal(c.penetration, jp.repeat(0.01, 2))
    np.testing.assert_array_almost_equal(
        c.pos, jp.array([[0.0, 0.0, 0.495]] * 2)
    )
    np.testing.assert_array_almost_equal(
        c.normal, jp.array([[0, 0.0, 1.0]] * 2), 5
    )


class CapsuleTest(absltest.TestCase):
  """Tests the capsule contact functions."""

  _CAP_CAP = """
    <mujoco model="two_capsules">
      <worldbody>
        <body name="body1" pos="0 0 0">
          <joint axis="1 0 0" name="free1" pos="0 0 0" type="free"/>
          <geom name="cap1" fromto="0.62235904  0.58846647 0.651046 1.5330081 0.33564585 0.977849" pos="0 0 0" size="0.05" type="capsule"/>
        </body>
        <body name="body2" pos="0 0 0">
          <joint axis="1 0 0" name="free2" pos="0 0 0" type="free"/>
          <geom name="cap2" fromto="0.5505271 0.60345304 0.476661 1.3900293 0.30709633 0.932082" pos="0 0 0" size="0.05" type="capsule"/>
        </body>
      </worldbody>
      <contact>
        <pair geom1="cap1" geom2="cap2"/>
      </contact>
    </mujoco>
  """

  def test_capsule_capsule(self):
    """Tests that two capsules are colliding."""
    sys = mjcf.loads(self._CAP_CAP)
    x, _ = kinematics.forward(sys, sys.init_q, jp.zeros(sys.qd_size()))
    c = geometry.contact(sys, x).take(0)
    self.assertGreaterEqual(c.penetration, 0)

  _PARALLEL_CAP = """
    <mujoco model="parallel_capsules">
      <worldbody>
        <body name="body1" pos="0 0 0">
          <joint axis="1 0 0" name="free1" pos="0 0 0" type="free"/>
          <geom name="cap1" fromto="-0.5 0.1 0.25 0.5 0.1 0.25" pos="0 0 0" size="0.1" type="capsule"/>
        </body>
        <body name="body2" pos="0 0 0">
          <joint axis="1 0 0" name="free2" pos="0 0 0" type="free"/>
          <geom name="cap2" fromto="-0.5 0.1 0.1 0.5 0.1 0.1" pos="0 0 0" size="0.1" type="capsule"/>
        </body>
      </worldbody>
      <contact>
        <pair geom1="cap1" geom2="cap2"/>
      </contact>
    </mujoco>
  """

  def test_parallel_capsules(self):
    """Tests that two parallel capsules are colliding at the midpoint."""
    sys = mjcf.loads(self._PARALLEL_CAP)
    x, _ = kinematics.forward(sys, sys.init_q, jp.zeros(sys.qd_size()))
    c = geometry.contact(sys, x).take(0)

    np.testing.assert_array_almost_equal(c.penetration, 0.05)
    np.testing.assert_array_almost_equal(
        c.pos, jp.array([0.0, 0.1, (0.15 + 0.2) / 2.0])
    )
    np.testing.assert_array_almost_equal(c.normal, jp.array([0, 0.0, 1.0]), 5)

  _CAP_BOX = """
    <mujoco model="cap_box">
      <worldbody>
        <body name="body1" pos="0 0 0.54">
          <joint axis="1 0 0" name="free1" pos="0 0 0" type="free"/>
          <geom name="cap1" fromto="-0.5 0 0 0.5 0 0" pos="0 0 0" size="0.05" type="capsule"/>
        </body>
        <body name="body2" pos="0 0 0" euler="0 0 0">
          <joint axis="1 0 0" name="free2" pos="0 0 0" type="free"/>
          <geom name="box" pos="0 0 0" size="0.5 0.5 0.5" type="box"/>
        </body>
      </worldbody>
    </mujoco>
  """

  def test_capsule_mesh(self):
    sys = mjcf.loads(self._CAP_BOX)
    x, _ = kinematics.forward(sys, sys.init_q, jp.zeros(sys.qd_size()))
    c = geometry.contact(sys, x)

    np.testing.assert_array_almost_equal((c.penetration > 0).sum(), 4)
    c = c.take(c.penetration.argsort()[-4:])
    np.testing.assert_array_almost_equal(c.penetration, jp.repeat(0.01, 4))
    np.testing.assert_array_almost_equal(c.pos[:, 2], jp.repeat(0.495, 4))
    np.testing.assert_array_almost_equal(
        c.normal, jp.array([[0, 0.0, 1.0]] * 4), 4
    )


class MeshTest(absltest.TestCase):
  """Tests the mesh contact functions."""

  _BOX_PLANE = """
    <mujoco model="box_plane">
      <worldbody>
        <geom name="floor" pos="0 0 0" size="40 40 40" type="plane"/>
        <body name="body1" pos="0 0 0.7" euler="45 0 0">
          <joint axis="1 0 0" name="free1" pos="0 0 0" type="free"/>
          <geom name="box" pos="0 0 0" size="0.5 0.5 0.5" type="box"/>
        </body>
      </worldbody>
    </mujoco>
  """

  def test_box_plane(self):
    sys = mjcf.loads(self._BOX_PLANE)
    x, _ = kinematics.forward(sys, sys.init_q, jp.zeros(sys.qd_size()))
    c = geometry.contact(sys, x)

    self.assertEqual(c.pos.shape[0], 8)  # 8 box corners
    self.assertEqual((c.penetration > 0.0).sum(), 2)  # 2 corners penetrating
    np.testing.assert_array_almost_equal(
        c.pos[c.penetration > 0.0].sort(axis=0),
        jp.array([[-0.5, 0.0, -0.00710678], [0.5, 0.0, -0.00710678]]),
    )
    self.assertTrue((c.normal == jp.array([0.0, 0.0, 1.0])).all())


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Observation functions.

get_obs_dict() supports modular observation specs, with observer=
  an Observer(): get the specified observation
    e.g.1. SimObserver(sim_datatype='body', sim_datacomp='vel',
      sim_dataname='root')
    e.g.2. LambdaObserver(observers=[obs1, obs2], fn=lambda x, y: x-y)
  'qp': includes all body info (pos, rot, vel, ang) of the component
  'root_joints': includes root body info and all joint info (pos, vel) of the
    component
  'root_z_joints': the same as 'root_joints' but remove root body's pos[:2]

index_preprocess() converts special specs, e.g. ('key1', 2), into int indices
  for index_obs()

index_obs() allows indexing with a list of indices
"""
import abc
import collections
from typing import Any, Dict, List, Tuple, Union
import brax
from brax.envs import Env
from brax.experimental.braxlines.common import sim_utils
from brax.experimental.composer import component_editor
from brax.experimental.composer import composer
from jax import numpy as jnp


class Observer(abc.ABC):
  """Observer."""

  def __init__(self, name: str = None, indices: Tuple[int] = None):
    assert name
    self.name = name
    if isinstance(indices, int):
      indices = (indices,)
    self.indices = indices
    self.initialized = False

  def initialize(self, sys):
    del sys
    self.initialized = True

  def index_obs(self, obs: jnp.ndarray):
    if self.indices is not None:
      obs = obs[..., self.indices]
    return obs

  def get_obs(self, sys, qp: brax.QP, info: brax.Info,
              cached_obs_dict: Dict[str, jnp.ndarray], component: Dict[str,
                                                                       Any]):
    if not self.initialized:
      self.initialize(sys)
    if self.name in cached_obs_dict:
      return cached_obs_dict[self.name].copy()
    obs = self._get_obs(sys, qp, info, cached_obs_dict, component)
    obs = self.index_obs(obs)
    return obs

  @abc.abstractmethod
  def _get_obs(self, sys, qp: brax.QP, info: brax.Info,
               cached_obs_dict: Dict[str, jnp.ndarray], component: Dict[str,
                                                                        Any]):
    raise NotImplementedError

  def __str__(self) -> str:
    return self.name

  def __repr__(self) -> str:
    return self.name


LAMBDA_FN_MAPPING = {
    '-': lambda x, y: x - y,
}


class LambdaObserver(Observer):
  """LambdaObserver."""

  def __init__(self, observers: List[Observer], fn, **kwargs):
    self.observers = observers
    if isinstance(fn, str):
      fn = LAMBDA_FN_MAPPING[fn]
    self.fn = fn
    super().__init__(**kwargs)

  def initialize(self, sys):
    for o in self.observers:
      o.initialize(sys)
    super().initialize(sys)

  def _get_obs(self, sys, qp: brax.QP, info: brax.Info,
               cached_obs_dict: Dict[str, jnp.ndarray], component: Dict[str,
                                                                        Any]):
    obses = [
        o.get_obs(sys, qp, info, cached_obs_dict, component)
        for o in self.observers
    ]
    return self.fn(*obses)


class SimObserver(Observer):
  """SimObserver."""

  def __init__(self,
               sdtype: str = 'body',
               sdcomp: str = 'pos',
               sdname: str = '',
               comp_name: str = '',
               name: str = None,
               indices: Tuple[int, ...] = None,
               **kwargs):
    sdname = component_editor.concat_name(sdname, comp_name)
    if not name:
      name = f'{sdtype}_{sdcomp}:{sdname}'
      if indices:
        name = f'{name}[{indices}]'
    self.sdtype = sdtype
    self.sdcomp = sdcomp
    self.sdname = sdname
    super().__init__(name=name, indices=indices, **kwargs)

  def initialize(self, sys):
    self.sim_indices, self.sim_info, self.sim_mask = sim_utils.names2indices(
        sys.config, names=[self.sdname], datatype=self.sdtype)
    self.sim_indices = self.sim_indices[0]  # list -> int
    super().initialize(sys)

  def _get_obs(self,
               sys,
               qp: brax.QP,
               info: brax.Info,
               cached_obs_dict: Dict[str, jnp.ndarray],
               component: Dict[str, Any] = None):
    """Get observation."""
    if self.sdtype == 'body':
      assert self.sdcomp in ('pos', 'rot', 'ang', 'vel'), self.sdcomp
      obs = getattr(qp, self.sdcomp)[self.sim_indices]
    elif self.sdtype == 'joint':
      joint_obs_dict = sim_utils.get_joint_value(sys, qp, self.sim_info)
      obs = list(joint_obs_dict.values())[0]
    elif self.sdtype == 'contact':
      assert self.sdcomp in ('vel', 'ang'), self.sdcomp
      v = getattr(info.contact, self.sdcomp)[self.sim_indices]
      v = jnp.clip(v, -1, 1)
      obs = jnp.reshape(v, v.shape[:-2] + (-1,))
    else:
      raise NotImplementedError(self.sdtype)
    return obs


def index_preprocess(indices: List[Any], env: Env = None) -> List[int]:
  """Preprocess indices to a list of ints and a list of str labels."""
  if indices is None:
    return None
  int_indices = []
  labels = []
  for index in indices:
    if isinstance(index, int):
      int_indices += [index]
      labels += [f'obs[{index}]']
    elif isinstance(index, tuple):
      assert len(index) == 2, 'tuple indexing is of form: (obs_dict_key, index)'
      key, i = index
      assert isinstance(env, composer.ComponentEnv), env
      assert env.observation_size  # ensure env.observer_shapes is set
      obs_shape = env.observer_shapes
      assert key in obs_shape, f'{key} not in {tuple(obs_shape.keys())}'
      int_indices += [obs_shape[key]['start'] + i]
      labels += [f'{key}[{i}]']
    else:
      raise NotImplementedError(index)
  return int_indices, labels


def index_obs(obs: jnp.ndarray, indices: List[Any], env: Env = None):
  """Index observation vector."""
  int_indices = index_preprocess(indices, env)
  return obs.take(int_indices, axis=-1)


def initialize_observers(observers: List[Union[Observer, str]], sys):
  """Initialize observers."""
  for o in observers:
    if isinstance(o, Observer):
      o.initialize(sys)


STRING_OBSERVERS = ('qp', 'root_joints', 'root_z_joints', 'cfrc')


def get_obs_dict(sys, qp: brax.QP, info: brax.Info, observer: Union[str,
                                                                    Observer],
                 cached_obs_dict: Dict[str, jnp.ndarray], component: Dict[str,
                                                                          Any]):
  """Observe."""
  obs_dict = collections.OrderedDict()
  if isinstance(observer, Observer):
    obs_dict[observer.name] = observer.get_obs(sys, qp, info, cached_obs_dict,
                                               component)
  elif observer == 'qp':
    # get all positions/orientations/velocities/ang velocities of all bodies
    bodies = component['bodies']
    indices = sim_utils.names2indices(sys.config, bodies, 'body')[0]
    for type_ in ('pos', 'rot', 'vel', 'ang'):
      for index, b in zip(indices, bodies):
        v = getattr(qp, type_)[index]
        key = f'body_{type_}:{b}'
        obs_dict[key] = v
  elif observer in ('root_joints', 'root_z_joints'):
    # get all positions/orientations/velocities/ang velocities of root bodies
    root = component['root']
    index = sim_utils.names2indices(sys.config, root, 'body')[0][0]
    for type_ in ('pos', 'rot', 'vel', 'ang'):
      v = getattr(qp, type_)[index]
      if observer == 'root_z_joints' and type_ == 'pos':
        # remove xy position
        v = v[2:]
      obs_dict[f'body_{type_}:{root}'] = v
    # get all joints
    joints = component['joints']
    _, joint_info, _ = sim_utils.names2indices(sys.config, joints, 'joint')
    joint_obs_dict = sim_utils.get_joint_value(sys, qp, joint_info)
    obs_dict = collections.OrderedDict(
        list(obs_dict.items()) + list(joint_obs_dict.items()))
  elif observer == 'cfrc':
    # external contact forces:
    # delta velocity (3,), delta ang (3,) * N bodies in the system
    bodies = component['bodies']
    indices = sim_utils.names2indices(sys.config, bodies, 'body')[0]
    for i, b in zip(indices, bodies):
      for type_ in ('vel', 'ang'):
        v = getattr(info.contact, type_)[i]
        v = jnp.clip(v, -1, 1)
        v = jnp.reshape(v, v.shape[:-2] + (-1,))
        key = f'contact_{type_}:{b}'
        obs_dict[key] = v
  else:
    raise NotImplementedError(observer)
  return obs_dict


def get_component_observers(component: Dict[str, Any],
                            observer_type: str = 'qp',
                            **observer_kwargs):
  """Get component-based Observers."""
  del component, observer_kwargs
  raise NotImplementedError(observer_type)


def get_edge_observers(component1: Dict[str, Any],
                       component2: Dict[str, Any],
                       observer_type: str = 'root_vec',
                       **observer_kwargs):
  """Get edge-based Observers."""
  if observer_type == 'root_vec':
    root1 = component1['root']
    root2 = component2['root']
    return LambdaObserver(
        name=f'dist__{root1}__{root2}',
        fn='-',
        observers=[
            SimObserver(sdname=component1['root'], **observer_kwargs),
            SimObserver(sdname=component2['root'], **observer_kwargs)
        ],
    )
  else:
    raise NotImplementedError(observer_type)



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Common default parameters."""
import copy

DEFAULT_PPO_PARAMS = {
    'ant':
        dict(
            num_timesteps=30000000,
            log_frequency=20,
            reward_scaling=10,
            episode_length=1000,
            normalize_observations=True,
            action_repeat=1,
            unroll_length=5,
            num_minibatches=32,
            num_update_epochs=4,
            discounting=0.97,
            learning_rate=3e-4,
            entropy_cost=1e-2,
            num_envs=2048,
            batch_size=1024),
    'humanoid':
        dict(
            num_timesteps=50000000,
            log_frequency=20,
            reward_scaling=0.1,
            episode_length=1000,
            normalize_observations=True,
            action_repeat=1,
            unroll_length=10,
            num_minibatches=32,
            num_update_epochs=8,
            discounting=0.97,
            learning_rate=3e-4,
            entropy_cost=1e-3,
            num_envs=2048,
            batch_size=1024),
    'fetch':
        dict(
            num_timesteps=100_000_000,
            log_frequency=20,
            reward_scaling=5,
            episode_length=1000,
            normalize_observations=True,
            action_repeat=1,
            unroll_length=20,
            num_minibatches=32,
            num_update_epochs=4,
            discounting=0.997,
            learning_rate=3e-4,
            entropy_cost=0.001,
            num_envs=2048,
            batch_size=256),
    'grasp':
        dict(
            num_timesteps=600_000_000,
            log_frequency=10,
            reward_scaling=10,
            episode_length=1000,
            normalize_observations=True,
            action_repeat=1,
            unroll_length=20,
            num_minibatches=32,
            num_update_epochs=2,
            discounting=0.99,
            learning_rate=3e-4,
            entropy_cost=0.001,
            num_envs=2048,
            batch_size=256),
    'halfcheetah':
        dict(
            num_timesteps=100_000_000,
            log_frequency=10,
            reward_scaling=1,
            episode_length=1000,
            normalize_observations=True,
            action_repeat=1,
            unroll_length=20,
            num_minibatches=32,
            num_update_epochs=8,
            discounting=0.95,
            learning_rate=3e-4,
            entropy_cost=0.001,
            num_envs=2048,
            batch_size=512),
    'ur5e':
        dict(
            num_timesteps=20_000_000,
            log_frequency=20,
            reward_scaling=10,
            episode_length=1000,
            normalize_observations=True,
            action_repeat=1,
            unroll_length=5,
            num_minibatches=32,
            num_update_epochs=4,
            discounting=0.95,
            learning_rate=2e-4,
            entropy_cost=1e-2,
            num_envs=2048,
            batch_size=1024,
            max_devices_per_host=8),
    'reacher':
        dict(
            num_timesteps=100_000_000,
            log_frequency=20,
            reward_scaling=5,
            episode_length=1000,
            normalize_observations=True,
            action_repeat=4,
            unroll_length=50,
            num_minibatches=32,
            num_update_epochs=8,
            discounting=0.95,
            learning_rate=3e-4,
            entropy_cost=1e-3,
            num_envs=2048,
            batch_size=256,
            max_devices_per_host=8),
}
DEFAULT_PPO_PARAMS.update(
    dict(
        hopper=DEFAULT_PPO_PARAMS['halfcheetah'],
        walker2d=DEFAULT_PPO_PARAMS['halfcheetah']))


def get_ppo_params(env_name: str,
                   timesteps_multiplier: float = 1,
                   default: str = None,
                   **kwargs):
  """Get Brax Training default ppo params."""
  if default:
    ppo_params = DEFAULT_PPO_PARAMS.get(env_name, DEFAULT_PPO_PARAMS[default])
  else:
    ppo_params = DEFAULT_PPO_PARAMS[env_name]
  ppo_params = copy.deepcopy(ppo_params)
  ppo_params['num_timesteps'] = int(ppo_params['num_timesteps'] *
                                    timesteps_multiplier)
  ppo_params.update(kwargs)
  return ppo_params



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Trains an ant to run in the +x direction."""

from brax.v2 import base
from brax.v2 import math
from brax.v2.envs import env
from brax.v2.io import mjcf
from etils import epath
import jax
from jax import numpy as jp


class Ant(env.PipelineEnv):



  # pyformat: disable
  """
  ### Description

  This environment is based on the environment introduced by Schulman, Moritz,
  Levine, Jordan and Abbeel in
  ["High-Dimensional Continuous Control Using Generalized Advantage Estimation"](https://arxiv.org/abs/1506.02438).

  The ant is a 3D robot consisting of one torso (free rotational body) with four
  legs attached to it with each leg having two links.

  The goal is to coordinate the four legs to move in the forward (right)
  direction by applying torques on the eight hinges connecting the two links of
  each leg and the torso (nine parts and eight hinges).

  ### Action Space

  The agent take a 8-element vector for actions.

  The action space is a continuous `(action, action, action, action, action,
  action, action, action)` all in `[-1, 1]`, where `action` represents the
  numerical torques applied at the hinge joints.

  | Num | Action                                                             | Control Min | Control Max | Name (in corresponding config)   | Joint | Unit         |
  |-----|--------------------------------------------------------------------|-------------|-------------|----------------------------------|-------|--------------|
  | 0   | Torque applied on the rotor between the torso and front left hip   | -1          | 1           | hip_1 (front_left_leg)           | hinge | torque (N m) |
  | 1   | Torque applied on the rotor between the front left two links       | -1          | 1           | ankle_1 (front_left_leg)         | hinge | torque (N m) |
  | 2   | Torque applied on the rotor between the torso and front right hip  | -1          | 1           | hip_2 (front_right_leg)          | hinge | torque (N m) |
  | 3   | Torque applied on the rotor between the front right two links      | -1          | 1           | ankle_2 (front_right_leg)        | hinge | torque (N m) |
  | 4   | Torque applied on the rotor between the torso and back left hip    | -1          | 1           | hip_3 (back_leg)                 | hinge | torque (N m) |
  | 5   | Torque applied on the rotor between the back left two links        | -1          | 1           | ankle_3 (back_leg)               | hinge | torque (N m) |
  | 6   | Torque applied on the rotor between the torso and back right hip   | -1          | 1           | hip_4 (right_back_leg)           | hinge | torque (N m) |
  | 7   | Torque applied on the rotor between the back right two links       | -1          | 1           | ankle_4 (right_back_leg)         | hinge | torque (N m) |

  ### Observation Space

  The state space consists of positional values of different body parts of the
  ant, followed by the velocities of those individual parts (their derivatives)
  with all the positions ordered before all the velocities.

  The observation is a `ndarray` with shape `(27,)` where the elements correspond to the following:

  | Num | Observation                                                  | Min  | Max | Name (in corresponding config)   | Joint | Unit                     |
  |-----|--------------------------------------------------------------|------|-----|----------------------------------|-------|--------------------------|
  | 0   | z-coordinate of the torso (centre)                           | -Inf | Inf | torso                            | free  | position (m)             |
  | 1   | w-orientation of the torso (centre)                          | -Inf | Inf | torso                            | free  | angle (rad)              |
  | 2   | x-orientation of the torso (centre)                          | -Inf | Inf | torso                            | free  | angle (rad)              |
  | 3   | y-orientation of the torso (centre)                          | -Inf | Inf | torso                            | free  | angle (rad)              |
  | 4   | z-orientation of the torso (centre)                          | -Inf | Inf | torso                            | free  | angle (rad)              |
  | 5   | angle between torso and first link on front left             | -Inf | Inf | hip_1 (front_left_leg)           | hinge | angle (rad)              |
  | 6   | angle between the two links on the front left                | -Inf | Inf | ankle_1 (front_left_leg)         | hinge | angle (rad)              |
  | 7   | angle between torso and first link on front right            | -Inf | Inf | hip_2 (front_right_leg)          | hinge | angle (rad)              |
  | 8   | angle between the two links on the front right               | -Inf | Inf | ankle_2 (front_right_leg)        | hinge | angle (rad)              |
  | 9   | angle between torso and first link on back left              | -Inf | Inf | hip_3 (back_leg)                 | hinge | angle (rad)              |
  | 10  | angle between the two links on the back left                 | -Inf | Inf | ankle_3 (back_leg)               | hinge | angle (rad)              |
  | 11  | angle between torso and first link on back right             | -Inf | Inf | hip_4 (right_back_leg)           | hinge | angle (rad)              |
  | 12  | angle between the two links on the back right                | -Inf | Inf | ankle_4 (right_back_leg)         | hinge | angle (rad)              |
  | 13  | x-coordinate velocity of the torso                           | -Inf | Inf | torso                            | free  | velocity (m/s)           |
  | 14  | y-coordinate velocity of the torso                           | -Inf | Inf | torso                            | free  | velocity (m/s)           |
  | 15  | z-coordinate velocity of the torso                           | -Inf | Inf | torso                            | free  | velocity (m/s)           |
  | 16  | x-coordinate angular velocity of the torso                   | -Inf | Inf | torso                            | free  | angular velocity (rad/s) |
  | 17  | y-coordinate angular velocity of the torso                   | -Inf | Inf | torso                            | free  | angular velocity (rad/s) |
  | 18  | z-coordinate angular velocity of the torso                   | -Inf | Inf | torso                            | free  | angular velocity (rad/s) |
  | 19  | angular velocity of angle between torso and front left link  | -Inf | Inf | hip_1 (front_left_leg)           | hinge | angle (rad)              |
  | 20  | angular velocity of the angle between front left links       | -Inf | Inf | ankle_1 (front_left_leg)         | hinge | angle (rad)              |
  | 21  | angular velocity of angle between torso and front right link | -Inf | Inf | hip_2 (front_right_leg)          | hinge | angle (rad)              |
  | 22  | angular velocity of the angle between front right links      | -Inf | Inf | ankle_2 (front_right_leg)        | hinge | angle (rad)              |
  | 23  | angular velocity of angle between torso and back left link   | -Inf | Inf | hip_3 (back_leg)                 | hinge | angle (rad)              |
  | 24  | angular velocity of the angle between back left links        | -Inf | Inf | ankle_3 (back_leg)               | hinge | angle (rad)              |
  | 25  | angular velocity of angle between torso and back right link  | -Inf | Inf | hip_4 (right_back_leg)           | hinge | angle (rad)              |
  | 26  | angular velocity of the angle between back right links       | -Inf | Inf | ankle_4 (right_back_leg)         | hinge | angle (rad)              |

  The (x,y,z) coordinates are translational DOFs while the orientations are
  rotational DOFs expressed as quaternions.

  ### Rewards

  The reward consists of three parts:

  - *reward_survive*: Every timestep that the ant is alive, it gets a reward of
    1.
  - *reward_forward*: A reward of moving forward which is measured as
    *(x-coordinate before action - x-coordinate after action)/dt*. *dt* is the
    time between actions - the default *dt = 0.05*. This reward would be
    positive if the ant moves forward (right) desired.
  - *reward_ctrl*: A negative reward for penalising the ant if it takes actions
    that are too large. It is measured as *coefficient **x**
    sum(action<sup>2</sup>)* where *coefficient* is a parameter set for the
    control and has a default value of 0.5.
  - *contact_cost*: A negative reward for penalising the ant if the external
    contact force is too large. It is calculated *0.5 * 0.001 *
    sum(clip(external contact force to [-1,1])<sup>2</sup>)*.

  ### Starting State

  All observations start in state (0.0, 0.0,  0.75, 1.0, 0.0  ... 0.0) with a
  uniform noise in the range of [-0.1, 0.1] added to the positional values and
  standard normal noise with 0 mean and 0.1 standard deviation added to the
  velocity values for stochasticity.

  Note that the initial z coordinate is intentionally selected to be slightly
  high, thereby indicating a standing up ant. The initial orientation is
  designed to make it face forward as well.

  ### Episode Termination

  The episode terminates when any of the following happens:

  1. The episode duration reaches a 1000 timesteps
  2. The y-orientation (index 2) in the state is **not** in the range
     `[0.2, 1.0]`

  ### Arguments

  No additional arguments are currently supported (in v2 and lower), but
  modifications can be made to the XML file in the assets folder (or by changing
  the path to a modified XML file in another folder).

  ```
  env = gym.make('Ant-v2')
  ```

  v3, v4, and v5 take gym.make kwargs such as ctrl_cost_weight,
  reset_noise_scale etc.

  ```
  env = gym.make('Ant-v5', ctrl_cost_weight=0.1, ....)
  ```

  ### Version History

  * v5: ported to Brax.
  * v4: all mujoco environments now use the mujoco bindings in mujoco>=2.1.3
  * v3: support for gym.make kwargs such as xml_file, ctrl_cost_weight,
        reset_noise_scale etc. rgb rendering comes from tracking camera (so
        agent does not run away from screen)
  * v2: All continuous control environments now use mujoco_py >= 1.50
  * v1: max_time_steps raised to 1000 for robot based tasks. Added
        reward_threshold to environments.
  * v0: Initial versions release (1.0.0)
  """
  # pyformat: enable


  def __init__(
      self,
      ctrl_cost_weight=0.5,
      use_contact_forces=False,
      contact_cost_weight=5e-4,
      healthy_reward=1.0,
      terminate_when_unhealthy=True,
      healthy_z_range=(0.2, 1.0),
      contact_force_range=(-1.0, 1.0),
      reset_noise_scale=0.1,
      exclude_current_positions_from_observation=True,
      backend='generalized',
      n_frames=5,
      **kwargs,
  ):
    path = epath.resource_path('brax') / 'v2/envs/assets/ant.xml'
    sys = mjcf.load(path)

    if backend == 'spring':
      sys = self.fix_spring_sys(sys)
      n_frames = 10
    super().__init__(sys=sys, backend=backend, n_frames=n_frames, **kwargs)

    self._ctrl_cost_weight = ctrl_cost_weight
    self._use_contact_forces = use_contact_forces
    self._contact_cost_weight = contact_cost_weight
    self._healthy_reward = healthy_reward
    self._terminate_when_unhealthy = terminate_when_unhealthy
    self._healthy_z_range = healthy_z_range
    self._contact_force_range = contact_force_range
    self._reset_noise_scale = reset_noise_scale
    self._exclude_current_positions_from_observation = (
        exclude_current_positions_from_observation
    )

    if self._use_contact_forces:
      raise NotImplementedError('use_contact_forces not implemented.')

  def reset(self, rng: jp.ndarray) -> env.State:
    """Resets the environment to an initial state."""
    rng, rng1, rng2 = jax.random.split(rng, 3)

    low, hi = -self._reset_noise_scale, self._reset_noise_scale
    q = self.sys.init_q + jax.random.uniform(
        rng1, (self.sys.q_size(),), minval=low, maxval=hi
    )
    qd = hi * jax.random.normal(rng2, (self.sys.qd_size(),))

    pipeline_state = self._pipeline.init(self.sys, q, qd)
    obs = self._get_obs(pipeline_state)

    reward, done, zero = jp.zeros(3)
    metrics = {
        'reward_forward': zero,
        'reward_survive': zero,
        'reward_ctrl': zero,
        'reward_contact': zero,
        'x_position': zero,
        'y_position': zero,
        'distance_from_origin': zero,
        'x_velocity': zero,
        'y_velocity': zero,
        'forward_reward': zero,
    }
    return env.State(pipeline_state, obs, reward, done, metrics)

  def step(self, state: env.State, action: jp.ndarray) -> env.State:
    """Run one timestep of the environment's dynamics."""
    pipeline_state0 = state.pipeline_state
    pipeline_state = self.pipeline_step(pipeline_state0, action)

    velocity = (pipeline_state.x.pos[0] - pipeline_state0.x.pos[0]) / self.dt
    forward_reward = velocity[0]

    min_z, max_z = self._healthy_z_range
    is_healthy = jp.where(pipeline_state.x.pos[0, 2] < min_z, x=0.0, y=1.0)
    is_healthy = jp.where(
        pipeline_state.x.pos[0, 2] > max_z, x=0.0, y=is_healthy
    )
    if self._terminate_when_unhealthy:
      healthy_reward = self._healthy_reward
    else:
      healthy_reward = self._healthy_reward * is_healthy
    ctrl_cost = self._ctrl_cost_weight * jp.sum(jp.square(action))
    contact_cost = 0.0

    obs = self._get_obs(pipeline_state)
    reward = forward_reward + healthy_reward - ctrl_cost - contact_cost
    done = 1.0 - is_healthy if self._terminate_when_unhealthy else 0.0
    state.metrics.update(
        reward_forward=forward_reward,
        reward_survive=healthy_reward,
        reward_ctrl=-ctrl_cost,
        reward_contact=-contact_cost,
        x_position=pipeline_state.x.pos[0, 0],
        y_position=pipeline_state.x.pos[0, 1],
        distance_from_origin=math.safe_norm(pipeline_state.x.pos[0]),
        x_velocity=velocity[0],
        y_velocity=velocity[1],
        forward_reward=forward_reward,
    )
    return state.replace(
        pipeline_state=pipeline_state, obs=obs, reward=reward, done=done
    )

  def _get_obs(self, pipeline_state: base.State) -> jp.ndarray:
    """Observe ant body position and velocities."""
    qpos = pipeline_state.q
    qvel = pipeline_state.qd

    if self._exclude_current_positions_from_observation:
      qpos = pipeline_state.q[2:]

    return jp.concatenate([qpos] + [qvel])

  def fix_spring_sys(self, sys):
    # TODO: add to custom fields in ant xml
    sys = sys.replace(
        link=sys.link.replace(
            inertia=sys.link.inertia.replace(
                i=jp.array([1.0 * jp.eye(3)] * sys.num_links()),
                mass=jp.ones_like(sys.link.inertia.mass),
            )
        )
    )
    link = sys.link.replace(constraint_limit_stiffness=jp.array([1_000.0] * 9))
    link = link.replace(constraint_stiffness=jp.array([4_000.0] * 9))
    link = link.replace(constraint_ang_damping=jp.array([10.0] * 9))
    link = link.replace(constraint_damping=jp.array([20.0] * 9))
    sys = sys.replace(link=link)
    sys = sys.replace(ang_damping=0.0)
    sys = sys.replace(dt=0.005)
    sys = sys.replace(actuator=sys.actuator.replace(gear=jp.array(8 * [100.0])))
    return sys



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Actuation of joints."""

import abc
from typing import List, Tuple

from brax import jumpy as jp
from brax import pytree
from brax.physics import config_pb2
from brax.physics import joints
from brax.physics.base import P, QP


class Actuator(abc.ABC):
  """Applies a torque to a joint."""

  __pytree_ignore__ = ('act_index', 'act_mask')

  def __init__(self, joint: joints.Joint, actuators: List[config_pb2.Actuator],
               act_index: List[Tuple[int, int]]):
    """Creates an actuator that applies torque to a joint given an act array.

    Args:
      joint: (batched) joint for this actuator to act upon
      actuators: list of actuators (all of the same type) to batch together
      act_index: indices from the act array that drive this Actuator
    """

    self.joint = jp.take(joint, [joint.index[a.joint] for a in actuators])
    self.strength = jp.array([a.strength for a in actuators])
    self.act_index = jp.array(act_index)
    self.act_mask = jp.where(self.act_index >= 0, 1., 0.)

  @abc.abstractmethod
  def apply_reduced(self, act: jp.ndarray, qp_p: QP, qp_c: QP) -> Tuple[P, P]:
    """Returns calculated impulses in reduced joint space."""

  def apply(self, qp: QP, act: jp.ndarray) -> P:
    """Applies torque to a joint.

    Args:
      qp: State data for system
      act: User-defined target for the actuator

    Returns:
      dP: The impulses that drive a joint
    """
    qp_p = jp.take(qp, self.joint.body_p.idx)
    qp_c = jp.take(qp, self.joint.body_c.idx)
    act = jp.take(act, self.act_index) * self.act_mask
    dang_p, dang_c = jp.vmap(type(self).apply_reduced)(self, act, qp_p, qp_c)

    # sum together all impulse contributions across parents and children
    body_idx = jp.concatenate((self.joint.body_p.idx, self.joint.body_c.idx))
    dp_ang = jp.concatenate((dang_p, dang_c))
    dp_ang = jp.segment_sum(dp_ang, body_idx, qp.pos.shape[0])

    return P(vel=jp.zeros_like(qp.vel), ang=dp_ang)


@pytree.register
class Angle(Actuator):
  """Applies torque to satisfy a target angle of a joint."""

  def apply_reduced(self, act: jp.ndarray, qp_p: QP, qp_c: QP) -> Tuple[P, P]:
    axis, angle = self.joint.axis_angle(qp_p, qp_c)
    axis, angle = jp.array(axis), jp.array(angle)

    limit_min, limit_max = self.joint.limit[:, 0], self.joint.limit[:, 1]
    act = jp.clip(act * jp.pi / 180, limit_min, limit_max)
    torque = (act - angle) * self.strength
    torque = jp.sum(jp.vmap(jp.multiply)(axis, torque), axis=0)

    dang_p = -self.joint.body_p.inertia * torque
    dang_c = self.joint.body_c.inertia * torque

    return dang_p, dang_c


@pytree.register
class Torque(Actuator):
  """Applies a direct torque to a joint."""

  def apply_reduced(self, act: jp.ndarray, qp_p: QP, qp_c: QP) -> Tuple[P, P]:
    axis, angle = self.joint.axis_angle(qp_p, qp_c)
    axis, angle = jp.array(axis), jp.array(angle)

    # clip torque if outside joint angle limits
    # * -1. so that positive actuation increases angle between parent and child
    torque = act * self.strength * -1.
    torque = jp.where(angle < self.joint.limit[:, 0], 0, torque)
    torque = jp.where(angle > self.joint.limit[:, 1], 0, torque)
    torque = jp.sum(jp.vmap(jp.multiply)(axis, torque), axis=0)

    dang_p = self.joint.body_p.inertia * torque
    dang_c = -self.joint.body_c.inertia * torque

    return dang_p, dang_c


def get(config: config_pb2.Config,
        all_joints: List[joints.Joint]) -> List[Actuator]:
  """Creates all actuators given a config and joints."""
  actuators = {}
  current_index = 0
  for actuator in config.actuators:
    joint = [j for j in all_joints if actuator.joint in j.index]
    if not joint:
      raise RuntimeError(f'joint not found: {actuator.joint}')
    joint = joint[0]
    joint_idx = joint.index[actuator.joint]
    if joint.free_dofs is not None:
      # padding out act index with dummy indices
      # e.g., suppose there are three joints with 1, 2, and 3 active degrees
      # of freedom, which have all been sphericalized.  The action vector will
      # have length 6, but needs to be unpacked into something with shape
      # [3, 3], so that we vectorize over the three spherical joints.  thus, we
      # construct a new tuple of indices that looks like:
      # act_index:
      # ( [0, -1, -1],
      #   [1,  2, -1],
      #   [3,  4,  5], )
      # then, we can simply mask out the -1 indices at act time
      free_dofs = joint.free_dofs[joint_idx]
      act_index = tuple(i if i - current_index < free_dofs else -1
                        for i in range(current_index, current_index +
                                       joint.dof))
      current_index += joint.free_dofs[joint_idx]
    else:
      act_index = tuple(range(current_index, current_index + joint.dof))
      current_index += joint.dof

    key = (actuator.WhichOneof('type'), joint.dof, joint)
    if key not in actuators:
      actuators[key] = []
    actuators[key].append((actuator, act_index))

  # ensure stable order for actuator application: actuator type, then dof
  actuators = sorted(actuators.items(), key=lambda kv: kv[0])
  ret = []
  for (actuator, _, joint), act_config_index in actuators:
    act_config = [c for c, _ in act_config_index]
    act_index = [a for _, a in act_config_index]
    if actuator == 'torque':
      ret.append(Torque(joint, act_config, act_index))
    elif actuator == 'angle':
      ret.append(Angle(joint, act_config, act_index))
    else:
      raise RuntimeError(f'unknown actuator type: {actuator}')
  return ret



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test for Brax's Replay buffers."""

from typing import List

from absl.testing import absltest
from absl.testing import parameterized
from brax.training import replay_buffers
import jax
import jax.numpy as jnp


def get_dummy_data():
  return {'a': jnp.zeros(()), 'b': jnp.zeros((5, 5))}


def get_dummy_batch(batch_size: int = 8):
  return {
      'a':
          jnp.arange(batch_size, dtype=jnp.float32),
      'b':
          jnp.reshape(
              jnp.arange(batch_size * 5 * 5, dtype=jnp.float32),
              (batch_size, 5, 5))
  }


def sorted_data(buffer_state: replay_buffers.ReplayBufferState) -> List[int]:
  """Returns the data of the buffer, rolled so set the current position to 0."""
  return jnp.sort(jnp.squeeze(buffer_state.data, axis=1)).tolist()


QUEUE_FACTORIES = [replay_buffers.UniformSamplingQueue, replay_buffers.Queue]


class QueueReplayTest(parameterized.TestCase):
  """Tests for Replay Buffers."""

  @parameterized.parameters(QUEUE_FACTORIES)
  def testInsert(self, queue_cls):
    replay_buffer = queue_cls(
        max_replay_size=10,
        dummy_data_sample=get_dummy_data(),
        sample_batch_size=2)
    rng = jax.random.PRNGKey(0)
    buffer_state = replay_buffer.init(rng)
    self.assertEqual(replay_buffer.size(buffer_state), 0)

    buffer_state = jax.jit(replay_buffer.insert)(buffer_state,
                                                 get_dummy_batch())
    self.assertEqual(replay_buffer.size(buffer_state), 8)

    # Hit the max replay_size.
    buffer_state = jax.jit(replay_buffer.insert)(buffer_state,
                                                 get_dummy_batch())
    self.assertEqual(replay_buffer.size(buffer_state), 10)

  @parameterized.parameters(QUEUE_FACTORIES)
  def testInvalidStateShape(self, queue_cls) -> None:
    max_replay_size = 10
    replay_buffer = queue_cls(
        max_replay_size=max_replay_size,
        dummy_data_sample=get_dummy_data(),
        sample_batch_size=2)
    rng = jax.random.PRNGKey(0)
    buffer_state = replay_buffer.init(rng)
    insert = jax.jit(replay_buffer.insert)

    # Make sure inserting in the buffer works wihtout crashing.
    insert(buffer_state, get_dummy_batch())

    # Expect an exception if `buffer_state.data` was corrupted.
    invalid_state = buffer_state.replace(data=jnp.arange(10))
    with self.assertRaises(ValueError) as context_manager:
      insert(invalid_state, get_dummy_batch())
    self.assertContainsSubsequence(str(context_manager.exception), 'shape')

    # Expect an exception if batch_size is larger than max_replay_size.
    insert(buffer_state, get_dummy_batch(max_replay_size))
    with self.assertRaises(ValueError) as context_manager:
      insert(buffer_state, get_dummy_batch(max_replay_size + 1))
    self.assertContainsSubsequence(
        str(context_manager.exception), 'maximum replay size')

  @parameterized.parameters(QUEUE_FACTORIES)
  def testInsertRollingBatches(self, queue_cls) -> None:
    max_size = 5
    replay_buffer = queue_cls(
        max_replay_size=max_size, dummy_data_sample=0, sample_batch_size=2)
    rng = jax.random.PRNGKey(0)
    buffer_state = replay_buffer.init(rng)
    self.assertEqual(replay_buffer.size(buffer_state), 0)
    self.assertEqual(sorted_data(buffer_state), [0, 0, 0, 0, 0])

    @jax.jit
    def insert(buffer_state, batch):
      return replay_buffer.insert(buffer_state, jnp.array(batch))

    buffer_state = insert(buffer_state, [1, 2, 3])
    self.assertEqual(replay_buffer.size(buffer_state), 3)
    self.assertEqual(sorted_data(buffer_state), [0, 0, 1, 2, 3])

    buffer_state = insert(buffer_state, [4, 5, 6])
    self.assertEqual(replay_buffer.size(buffer_state), 5)
    self.assertEqual(sorted_data(buffer_state), [2, 3, 4, 5, 6])

    buffer_state = insert(buffer_state, [7, 8, 9, 10, 11])
    self.assertEqual(replay_buffer.size(buffer_state), 5)
    self.assertEqual(sorted_data(buffer_state), [7, 8, 9, 10, 11])

  @parameterized.parameters(QUEUE_FACTORIES)
  def testInsertRolling2DimBatches(self, queue_cls) -> None:
    max_size = 5
    replay_buffer = queue_cls(
        max_replay_size=max_size,
        dummy_data_sample=jnp.zeros((2,)),
        sample_batch_size=2)
    rng = jax.random.PRNGKey(0)
    buffer_state = replay_buffer.init(rng)
    self.assertEqual(replay_buffer.size(buffer_state), 0)
    self.assertTrue(
        jnp.all(buffer_state.data == jnp.array([[0, 0], [0, 0], [0, 0], [0, 0],
                                                [0, 0]])))

    @jax.jit
    def insert(buffer_state, batch):
      return replay_buffer.insert(buffer_state,
                                  jnp.array(batch, dtype=jnp.float32))

    buffer_state = insert(buffer_state, [[1, 1], [2, 2], [3, 3]])
    self.assertEqual(replay_buffer.size(buffer_state), 3)
    self.assertTrue(
        jnp.all(buffer_state.data == jnp.array([[1, 1], [2, 2], [3, 3], [0, 0],
                                                [0, 0]])))

    buffer_state = insert(buffer_state, [[4, 4], [5, 5], [6, 6]])
    self.assertEqual(replay_buffer.size(buffer_state), 5)
    self.assertTrue(
        jnp.all(buffer_state.data == jnp.array([[2, 2], [3, 3], [4, 4], [5, 5],
                                                [6, 6]])))

  def testUniformSamplingQueueSample(self):
    replay_buffer = replay_buffers.UniformSamplingQueue(
        max_replay_size=10,
        dummy_data_sample=get_dummy_data(),
        sample_batch_size=2)
    rng = jax.random.PRNGKey(0)
    buffer_state = replay_buffer.init(rng)
    self.assertEqual(replay_buffer.size(buffer_state), 0)

    buffer_state = jax.jit(replay_buffer.insert)(buffer_state,
                                                 get_dummy_batch())
    self.assertEqual(replay_buffer.size(buffer_state), 8)

    buffer_state, samples = jax.jit(replay_buffer.sample)(buffer_state)
    self.assertEqual(replay_buffer.size(buffer_state), 8)
    self.assertEqual(samples['a'].shape, (2,))
    self.assertEqual(samples['b'].shape, (2, 5, 5))
    for sample in samples['b']:
      self.assertSequenceEqual(
          list(jnp.reshape(sample - sample[0, 0], (-1,))), range(5 * 5))

  def testUniformSamplingQueueCyclicSample(self):
    replay_buffer = replay_buffers.UniformSamplingQueue(
        max_replay_size=10, dummy_data_sample=0, sample_batch_size=2)
    rng = jax.random.PRNGKey(0)
    buffer_state = replay_buffer.init(rng)
    insert = jax.jit(replay_buffer.insert)

    self.assertEqual(replay_buffer.size(buffer_state), 0)

    buffer_state = insert(buffer_state, jnp.zeros(10, dtype=jnp.int32))
    self.assertEqual(replay_buffer.size(buffer_state), 10)
    self.assertEqual(buffer_state.current_position, 0)

  def testQueueSamplePyTree(self):
    replay_buffer = replay_buffers.Queue(
        max_replay_size=10,
        dummy_data_sample=get_dummy_data(),
        sample_batch_size=2)
    rng = jax.random.PRNGKey(0)
    buffer_state = replay_buffer.init(rng)
    self.assertEqual(replay_buffer.size(buffer_state), 0)

    buffer_state = jax.jit(replay_buffer.insert)(buffer_state,
                                                 get_dummy_batch())
    self.assertEqual(replay_buffer.size(buffer_state), 8)

    buffer_state, samples = jax.jit(replay_buffer.sample)(buffer_state)
    self.assertEqual(replay_buffer.size(buffer_state), 6)
    self.assertEqual(samples['a'].shape, (2,))
    self.assertEqual(samples['b'].shape, (2, 5, 5))
    for sample in samples['b']:
      self.assertSequenceEqual(
          list(jnp.reshape(sample - sample[0, 0], (-1,))), range(5 * 5))

  def testQueueSample(self):
    replay_buffer = replay_buffers.Queue(
        max_replay_size=10, dummy_data_sample=0, sample_batch_size=3)
    rng = jax.random.PRNGKey(0)

    buffer_state = replay_buffer.init(rng)
    insert = jax.jit(replay_buffer.insert)
    sample = jax.jit(replay_buffer.sample)

    buffer_state = insert(buffer_state, jnp.arange(4))
    self.assertSequenceStartsWith([0, 1, 2, 3], list(buffer_state.data))
    self.assertEqual(buffer_state.current_position, 4)
    self.assertEqual(replay_buffer.size(buffer_state), 4)

    buffer_state = insert(buffer_state, jnp.arange(4, 10))
    self.assertSequenceEqual(
        list(buffer_state.data), [0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
    self.assertEqual(buffer_state.current_position, 0)
    self.assertEqual(replay_buffer.size(buffer_state), 10)

    buffer_state, samples = sample(buffer_state)
    self.assertSequenceEqual(list(samples), [0, 1, 2])
    self.assertEqual(replay_buffer.size(buffer_state), 7)

    buffer_state, samples = sample(buffer_state)
    self.assertSequenceEqual(list(samples), [3, 4, 5])
    self.assertEqual(replay_buffer.size(buffer_state), 4)

    buffer_state, samples = sample(buffer_state)
    self.assertSequenceEqual(list(samples), [6, 7, 8])
    self.assertEqual(replay_buffer.size(buffer_state), 1)

    buffer_state = insert(buffer_state, jnp.arange(20, 25))
    buffer_state, samples = sample(buffer_state)
    self.assertSequenceEqual(list(samples), [9, 20, 21])
    self.assertEqual(replay_buffer.size(buffer_state), 3)

    buffer_state, samples = sample(buffer_state)
    self.assertSequenceEqual(list(samples), [22, 23, 24])
    self.assertEqual(replay_buffer.size(buffer_state), 0)

  def testQueueInsertWhenFull(self):
    replay_buffer = replay_buffers.Queue(
        max_replay_size=10, dummy_data_sample=0, sample_batch_size=3)
    rng = jax.random.PRNGKey(0)

    buffer_state = replay_buffer.init(rng)
    insert = jax.jit(replay_buffer.insert)

    buffer_state = insert(buffer_state, jnp.arange(10))
    buffer_state = insert(buffer_state, jnp.arange(10, 12))
    self.assertSequenceEqual(
        list(buffer_state.data), [10, 11, 2, 3, 4, 5, 6, 7, 8, 9])
    self.assertEqual(buffer_state.current_position, 2)
    self.assertEqual(replay_buffer.size(buffer_state), 10)

  def testQueueWrappedSample(self):
    replay_buffer = replay_buffers.Queue(
        max_replay_size=10, dummy_data_sample=0, sample_batch_size=7)
    rng = jax.random.PRNGKey(0)

    buffer_state = replay_buffer.init(rng)
    insert = jax.jit(replay_buffer.insert)
    sample = jax.jit(replay_buffer.sample)

    buffer_state = insert(buffer_state, jnp.arange(10))
    buffer_state = insert(buffer_state, jnp.arange(10, 16))
    self.assertSequenceEqual(
        list(buffer_state.data), [10, 11, 12, 13, 14, 15, 6, 7, 8, 9])
    self.assertEqual(replay_buffer.size(buffer_state), 10)
    self.assertEqual(buffer_state.current_position, 6)

    # This sample contains elements from both the beggining and the end of
    # the buffer.
    buffer_state, samples = sample(buffer_state)
    self.assertSequenceEqual(list(samples), [6, 7, 8, 9, 10, 11, 12])
    self.assertEqual(replay_buffer.size(buffer_state), 3)
    self.assertEqual(buffer_state.current_position, 6)

  def testQueueBatchSizeEqualsMaxSize(self):
    batch_size = 5
    replay_buffer = replay_buffers.Queue(
        max_replay_size=batch_size,
        dummy_data_sample=0,
        sample_batch_size=batch_size)
    rng = jax.random.PRNGKey(0)

    buffer_state = replay_buffer.init(rng)
    insert = jax.jit(replay_buffer.insert)
    sample = jax.jit(replay_buffer.sample)

    buffer_state = insert(buffer_state, jnp.arange(batch_size))
    buffer_state, samples = sample(buffer_state)
    self.assertSequenceEqual(list(samples), range(batch_size))
    self.assertEqual(buffer_state.current_size, 0)

    buffer_state = insert(buffer_state, jnp.zeros(batch_size, dtype=jnp.int32))
    buffer_state = insert(buffer_state, jnp.ones(batch_size, dtype=jnp.int32))
    buffer_state, samples = sample(buffer_state)
    self.assertSequenceEqual(list(samples), [1] * batch_size)
    self.assertEqual(buffer_state.current_size, 0)

  def testQueueSampleFromEmpty(self):
    batch_size = 5
    replay_buffer = replay_buffers.Queue(
        max_replay_size=batch_size,
        dummy_data_sample=0,
        sample_batch_size=batch_size)
    rng = jax.random.PRNGKey(0)

    buffer_state = replay_buffer.init(rng)
    insert = jax.jit(replay_buffer.insert)
    sample = jax.jit(replay_buffer.sample)

    buffer_state = insert(buffer_state, jnp.arange(batch_size))
    buffer_state, samples = sample(buffer_state)
    self.assertSequenceEqual(list(samples), range(batch_size))
    self.assertEqual(buffer_state.current_size, 0)

    buffer_state, samples = sample(buffer_state)
    self.assertSequenceEqual(list(samples), [0] * batch_size)
    self.assertEqual(buffer_state.current_size, 0)

    buffer_state = insert(buffer_state, jnp.arange(10, 13))
    self.assertEqual(buffer_state.current_size, 3)
    buffer_state, samples = sample(buffer_state)
    self.assertSequenceEqual(list(samples), [10, 11, 12, 0, 0])
    self.assertEqual(buffer_state.current_size, 0)

if __name__ == '__main__':
  absltest.main()



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Geometry functions and classes to be used for collision detection."""

import itertools
from typing import List, Mapping, Tuple

from brax import jumpy as jp
from brax import math
from brax import pytree
from brax.physics import bodies
from brax.physics import config_pb2
from brax.physics.base import QP, vec_to_arr

# Coordinates of the 8 corners of a box.
_BOX_CORNERS = jp.array(list(itertools.product((-1, 1), (-1, 1), (-1, 1))))

# pyformat: disable
# The faces of a triangulated box, i.e. the indices in _BOX_CORNERS of the
# vertices of the 12 triangles (two triangles for each side of the box).
_TRIANGULATED_BOX_FACES = [
    0, 4, 1, 4, 1, 5,  # front
    0, 4, 2, 2, 4, 6,  # bottom
    6, 4, 5, 6, 5, 7,  # right
    2, 6, 3, 3, 6, 7,  # back
    1, 3, 5, 5, 3, 7,  # top
    0, 2, 1, 1, 2, 3,  # left
]
# Normals of the triangulated box faces above.
_TRIANGULATED_BOX_FACE_NORMALS = [
    [0, -1., 0],  # front
    [0, -1., 0],
    [0, 0, -1.],  # bottom
    [0, 0, -1.],
    [+1., 0, 0],  # right
    [+1., 0, 0],
    [0, +1., 0],  # back
    [0, +1., 0],
    [0, 0, +1.],  # top
    [0, 0, +1.],
    [-1., 0, 0],  # left
    [-1., 0, 0],
]
# Polygon box faces using a clockwise winding order convention.
_BOX_FACES = [
    0, 1, 5, 4,  # front
    0, 4, 6, 2,  # bottom
    6, 4, 5, 7,  # right
    2, 6, 7, 3,  # back
    1, 3, 7, 5,  # top
    0, 2, 3, 1,  # left
]
# Normals of the polygon box faces above.
_BOX_FACE_NORMALS = [
    [0, -1., 0],  # front
    [0, 0, -1.],  # bottom
    [+1., 0, 0],  # right
    [0, +1., 0],  # back
    [0, 0, +1.],  # top
    [-1., 0, 0],  # left
]
# pyformat: enable


@pytree.register
class Collidable:
  """Part of a body (with geometry and mass/inertia) that can collide.

  Collidables can repeat for a geometry. e.g. a body with a box collider has 8
  corner collidables.
  """

  def __init__(self, collidables: List[config_pb2.Body], body: bodies.Body):
    self.body = jp.take(body, [body.index[c.name] for c in collidables])
    self.pos = jp.array(
        [vec_to_arr(c.colliders[0].position) for c in collidables])
    self.friction = jp.array(
        [c.colliders[0].material.friction for c in collidables])
    self.elasticity = jp.array(
        [c.colliders[0].material.elasticity for c in collidables])

  def position(self, qp: QP) -> math.Vector3:
    """Returns the collidable's position in world space."""
    pos = jp.take(qp.pos, self.body.idx)
    rot = jp.take(qp.rot, self.body.idx)
    return pos + jp.vmap(math.rotate)(self.pos, rot)


@pytree.register
class Contact:
  """Stores information about contacts between two collidables."""

  def __init__(self, pos: jp.ndarray, vel: jp.ndarray, normal: jp.ndarray,
               penetration: jp.ndarray):
    """Creates a Contact.

    Args:
      pos: contact position in world space
      vel: contact velocity in world space
      normal: normal vector of surface providing contact
      penetration: distance the two collidables are penetrating one-another
    """
    self.pos = pos
    self.vel = vel
    self.normal = normal
    self.penetration = penetration


@pytree.register
class Box(Collidable):
  """A box with corner collidables."""

  def __init__(self, boxes: List[config_pb2.Body], body: bodies.Body):
    super().__init__(boxes, body)
    corners = []
    for b in boxes:
      col = b.colliders[0]
      rot = math.euler_to_quat(vec_to_arr(col.rotation))
      box = _BOX_CORNERS * vec_to_arr(col.box.halfsize)
      box = jp.vmap(math.rotate, include=(True, False))(box, rot)
      box = box + vec_to_arr(col.position)
      corners.append(box)
    self.corner = jp.array(corners)


@pytree.register
class BaseMesh(Collidable):
  """Base class for mesh colliders."""

  def __init__(self, collidables: List[config_pb2.Body], body: bodies.Body,
               vertices: jp.ndarray, faces: jp.ndarray,
               face_normals: jp.ndarray):
    super().__init__(collidables, body)
    new_faces = []
    for face_normal, face in zip(face_normals, faces):
      # Fix the winding order of the faces if necessary by checking the first
      # two edges of each face for a clockwise winding order.
      winding_order = jp.vmap(jp.dot)(jp.cross(
          face[:, 0] - face[:, -1], face[:, 0] - face[:, 1]), face_normal) >= 0
      face = jp.vmap(lambda x, y: jp.where(x, y, y[::-1]))(winding_order, face)
      new_faces.append(face)
    self.faces = jp.array(new_faces)
    self.face_normals = face_normals
    self.vertices = vertices


@pytree.register
class BoxMesh(BaseMesh):
  """A box converted into a mesh."""

  def __init__(self, boxes: List[config_pb2.Body], body: bodies.Body,
               box_faces: List[int], box_face_normals: List[List[float]]):
    vertices = []
    faces = []
    face_normals = []
    num_face_points = len(box_faces) // len(box_face_normals)
    box_face_normals = jp.array(box_face_normals)
    for b in boxes:
      col = b.colliders[0]
      rot = math.euler_to_quat(vec_to_arr(col.rotation))
      vertex = _BOX_CORNERS * vec_to_arr(col.box.halfsize)
      vertex = jp.vmap(math.rotate, include=(True, False))(vertex, rot)
      vertex = vertex + vec_to_arr(col.position)
      vertices.append(vertex)

      # Apply rotation to face normals.
      face_normal = jp.vmap(
          math.rotate, include=(True, False))(box_face_normals, rot)
      face_normals.append(face_normal)

      # Each face consists of a `num_face_points` dimensional polygon.
      face = jp.reshape(jp.take(vertex, box_faces), (-1, num_face_points, 3))
      faces.append(face)

    super().__init__(boxes, body, jp.array(vertices), jp.array(faces),
                     jp.array(face_normals))


@pytree.register
class TriangulatedBox(BoxMesh):
  """A box converted into a triangular mesh."""

  def __init__(self, boxes: List[config_pb2.Body], body: bodies.Body):
    super().__init__(boxes, body, _TRIANGULATED_BOX_FACES,
                     _TRIANGULATED_BOX_FACE_NORMALS)


@pytree.register
class HullBox(BoxMesh):
  """A box converted into a polygon mesh."""

  def __init__(self, boxes: List[config_pb2.Body], body: bodies.Body):
    super().__init__(boxes, body, _BOX_FACES, _BOX_FACE_NORMALS)


@pytree.register
class Plane(Collidable):
  """An infinite plane with normal pointing in the +z direction."""


@pytree.register
class ClippedPlane(Collidable):
  """A clipped plane with the normal pointing in a user specified direction."""

  def  __init__(self, planes: List[config_pb2.Body], body: bodies.Body):
    super().__init__(planes, body)
    normals, xdir, ydir = [], [], []
    halfsize_x, halfsize_y, pos = [], [], []
    for p in planes:
      col = p.colliders[0]
      rot = math.euler_to_quat(vec_to_arr(col.rotation))
      normal = math.rotate(jp.array([0., 0., 1.]), rot)
      x = math.rotate(jp.array([1., 0., 0.]), rot)
      y = math.rotate(jp.array([0., 1., 0.]), rot)
      normals.append(normal)
      xdir.append(x)
      ydir.append(y)
      halfsize_x.append(col.clipped_plane.halfsize_x)
      halfsize_y.append(col.clipped_plane.halfsize_y)
      pos.append(vec_to_arr(col.position))

    self.normal = jp.array(normals)
    self.x, self.y = jp.array(xdir), jp.array(ydir)
    self.halfsize_x = jp.array(halfsize_x)
    self.halfsize_y = jp.array(halfsize_y)
    self.pos = jp.array(pos)


@pytree.register
class Capsule(Collidable):
  """A capsule with an ends pointing in the +z, -z directions."""

  def __init__(self, capsules: List[config_pb2.Body], body: bodies.Body):
    super().__init__(capsules, body)
    ends = []
    radii = []
    for c in capsules:
      col = c.colliders[0]
      axis = math.rotate(
          jp.array([0., 0., 1.]), math.euler_to_quat(vec_to_arr(col.rotation)))
      segment_length = col.capsule.length * 0.5 - col.capsule.radius
      ends.append(axis * segment_length)
      radii.append(col.capsule.radius)
    self.end = jp.array(ends)
    self.radius = jp.array(radii)


@pytree.register
class CapsuleEnd(Collidable):
  """A capsule with variable ends either in the +z or -z directions."""

  def __init__(self, capsules: List[config_pb2.Body], body: bodies.Body):
    super().__init__(capsules, body)
    ends = []
    radii = []
    for c in capsules:
      col = c.colliders[0]
      axis = math.rotate(
          jp.array([0., 0., 1.]), math.euler_to_quat(vec_to_arr(col.rotation)))
      segment_length = col.capsule.length * 0.5 - col.capsule.radius
      caps = []
      for end in [col.capsule.end] if col.capsule.end else [-1, 1]:
        caps.append(vec_to_arr(col.position) + end * axis * segment_length)
      ends.append(caps)
      radii.append(col.capsule.radius)
    # if there's a mix of 1 and 2 end capsules, pad the 1-end capsules
    # with a dummy cap that has 0 radius with the same end-location.
    # this facilitates vectorizing over the cap dimension
    if len(set([len(e) for e in ends])) != 1:
      for e in ends:
        if len(e) == 1:
          e.append(e[0])

    self.end = jp.array(ends)
    self.radius = jp.array(radii)


@pytree.register
class HeightMap(Collidable):
  """A height map with heights in a grid layout."""

  def __init__(self, heightmaps: List[config_pb2.Body], body: bodies.Body):
    super().__init__(heightmaps, body)
    heights = []
    cell_sizes = []
    for h in heightmaps:
      col = h.colliders[0]
      mesh_size = int(jp.sqrt(len(col.heightMap.data)))
      if len(col.heightMap.data) != mesh_size**2:
        raise ValueError('height map data length should be a perfect square.')
      height = jp.array(col.heightMap.data).reshape((mesh_size, mesh_size))
      heights.append(height)
      cell_sizes.append(col.heightMap.size / (mesh_size - 1))
    self.height = jp.array(heights)
    self.cell_size = jp.array(cell_sizes)


@pytree.register
class Mesh(BaseMesh):
  """A triangular mesh with vertex or face collidables."""

  def __init__(self, meshes: List[config_pb2.Body], body: bodies.Body,
               mesh_geoms: Mapping[str, config_pb2.MeshGeometry]):
    """Initializes a triangular mesh collider.

    Args:
      meshes: Mesh colliders of the body in the config.
      body: The body that the mesh colliders belong to.
      mesh_geoms: The dictionary of the mesh geometries keyed by their names.
    """
    geoms = [mesh_geoms[m.colliders[0].mesh.name] for m in meshes]

    vertices = []
    faces = []
    face_normals = []
    for m, g in zip(meshes, geoms):
      col = m.colliders[0]
      rot = math.euler_to_quat(vec_to_arr(col.rotation))
      scale = col.mesh.scale if col.mesh.scale else 1

      # Apply scaling and body transformations to the vertices.
      vertex = jp.array(
          [[v.x * scale, v.y * scale, v.z * scale] for v in g.vertices])
      vertex = jp.vmap(math.rotate, include=(True, False))(vertex, rot)
      vertex = vertex + vec_to_arr(col.position)
      vertices.append(vertex)

      # Each face is a triangle.
      face = jp.reshape(jp.take(vertex, g.faces), (-1, 3, 3))
      faces.append(face)

      # Apply rotation to face normals.
      face_normal = jp.array([vec_to_arr(n) for n in g.face_normals])
      face_normal = jp.vmap(
          math.rotate, include=(True, False))(face_normal, rot)
      face_normals.append(face_normal)

    super().__init__(meshes, body, jp.array(vertices), jp.array(faces),
                     jp.array(face_normals))


@pytree.register
class PointMesh(Mesh):
  """A triangular mesh with vertex collidables."""


def closest_segment_point(a: math.Vector3, b: math.Vector3,
                          pt: math.Vector3) -> math.Vector3:
  """Returns the closest point on the a-b line segment to a point pt."""
  ab = b - a
  t = jp.dot(pt - a, ab) / (jp.dot(ab, ab) + 1e-6)
  return a + jp.clip(t, 0., 1.) * ab


def closest_segment_point_and_dist(
    a: math.Vector3, b: math.Vector3,
    pt: math.Vector3) -> Tuple[math.Vector3, jp.ndarray]:
  """Returns the closest point and distance^2 on the a-b line segment to pt."""
  p = closest_segment_point(a, b, pt)
  v = pt - p
  return p, jp.dot(v, v)


def closest_segment_point_plane(a: math.Vector3, b: math.Vector3,
                                p0: math.Vector3,
                                plane_normal: math.Vector3) -> math.Vector3:
  """Gets the closest point between a line segment and a plane."""
  # If a line segment is parametrized as S(t) = a + t * (b - a), we can
  # plug it into the plane equation dot(n, S(t)) - d = 0, then solve for t to
  # get the line-plane intersection. We then clip t to be in [0, 1] to be on
  # the line segment.
  n = plane_normal
  d = jp.dot(p0, n)  # shortest distance from origin to plane
  denom = jp.dot(n, b - a)
  t = (d - jp.dot(n, a)) / (denom + 1e-6)
  t = jp.clip(t, 0, 1)
  segment_point = a + t * (b - a)
  return segment_point


def _closest_segment_to_segment_points(
    a0: math.Vector3, a1: math.Vector3, b0: math.Vector3,
    b1: math.Vector3) -> Tuple[math.Vector3, math.Vector3, float, float]:
  """Returns closest points on two line segments, and the barycentric vars."""
  # Gets the closest segment points by first finding the closest points
  # between two lines. Points are then clipped to be on the line segments
  # and edge cases with clipping are handled.
  dir_a = a1 - a0
  len_a = jp.safe_norm(dir_a)
  len_a += 1e-6 * (len_a == 0.)
  dir_a = dir_a / len_a
  half_len_a = len_a * 0.5

  dir_b = b1 - b0
  len_b = jp.safe_norm(dir_b)
  len_b += 1e-6 * (len_b == 0.)
  dir_b = dir_b / len_b
  half_len_b = len_b * 0.5

  # Segment mid-points.
  a_mid = a0 + dir_a * half_len_a
  b_mid = b0 + dir_b * half_len_b

  # Translation between two segment mid-points.
  trans = a_mid - b_mid

  # Parametrize points on each line as follows:
  #  point_on_a = a_mid + t_a * dir_a
  #  point_on_b = b_mid + t_b * dir_b
  # and analytically minimize the distance between the two points.
  dira_dot_dirb = dir_a.dot(dir_b)
  dira_dot_trans = dir_a.dot(trans)
  dirb_dot_trans = dir_b.dot(trans)
  denom = 1 - dira_dot_dirb * dira_dot_dirb

  orig_t_a = (-dira_dot_trans + dira_dot_dirb * dirb_dot_trans) / (denom + 1e-6)
  orig_t_b = dirb_dot_trans + orig_t_a * dira_dot_dirb
  t_a = jp.clip(orig_t_a, -half_len_a, half_len_a)
  t_b = jp.clip(orig_t_b, -half_len_b, half_len_b)

  best_a = a_mid + dir_a * t_a
  best_b = b_mid + dir_b * t_b

  # Resolve edge cases where both closest points are clipped to the segment
  # endpoints by recalculating the closest segment points for the current
  # clipped points, and then picking the pair of points with smallest
  # distance. An example of this edge case is when lines intersect but line
  # segments don't.
  new_a, d1 = closest_segment_point_and_dist(a0, a1, best_b)
  new_b, d2 = closest_segment_point_and_dist(b0, b1, best_a)
  best_a = jp.where(d1 < d2, new_a, best_a)
  best_b = jp.where(d1 < d2, best_b, new_b)

  # Transform the barycentric coordinate into [0, 1].
  t_a = (orig_t_a + half_len_a) / len_a
  t_b = (orig_t_b + half_len_b) / len_b

  return best_a, best_b, t_a, t_b


def closest_segment_to_segment_points(
    a0: math.Vector3, a1: math.Vector3, b0: math.Vector3,
    b1: math.Vector3) -> Tuple[math.Vector3, math.Vector3]:
  """Returns closest points on two line segments."""
  best_a, best_b, *_ = _closest_segment_to_segment_points(a0, a1, b0, b1)
  return best_a, best_b


def closest_triangle_point(p0: math.Vector3, p1: math.Vector3, p2: math.Vector3,
                           pt: math.Vector3) -> math.Vector3:
  """Gets the closest point on a triangle to another point in space."""
  # Parametrize the triangle s.t. a point inside the triangle is
  # Q = p0 + u * e0 + v * e1, when 0 <= u <= 1, 0 <= v <= 1, and
  # 0 <= u + v <= 1. Let e0 = (p1 - p0) and e1 = (p2 - p0).
  # We analytically minimize the distance between the point pt and Q.
  e0 = p1 - p0
  e1 = p2 - p0
  a = e0.dot(e0)
  b = e0.dot(e1)
  c = e1.dot(e1)
  d = pt - p0
  # The determinant is 0 only if the angle between e1 and e0 is 0
  # (i.e. the triangle has overlapping lines).
  det = (a * c - b * b)
  u = (c * e0.dot(d) - b * e1.dot(d)) / det
  v = (-b * e0.dot(d) + a * e1.dot(d)) / det
  inside = (0 <= u) & (u <= 1) & (0 <= v) & (v <= 1) & (u + v <= 1)
  closest_p = p0 + u * e0 + v * e1
  d0 = (closest_p - pt).dot(closest_p - pt)

  # If the closest point is outside the triangle, it must be on an edge, so we
  # check each triangle edge for a closest point to the point pt.
  closest_p1, d1 = closest_segment_point_and_dist(p0, p1, pt)
  closest_p = jp.where((d0 < d1) & inside, closest_p, closest_p1)
  min_d = jp.where((d0 < d1) & inside, d0, d1)

  closest_p2, d2 = closest_segment_point_and_dist(p1, p2, pt)
  closest_p = jp.where(d2 < min_d, closest_p2, closest_p)
  min_d = jp.minimum(min_d, d2)

  closest_p3, d3 = closest_segment_point_and_dist(p2, p0, pt)
  closest_p = jp.where(d3 < min_d, closest_p3, closest_p)
  min_d = jp.minimum(min_d, d3)

  return closest_p


def closest_segment_triangle_points(
    a: math.Vector3, b: math.Vector3, p0: math.Vector3, p1: math.Vector3,
    p2: math.Vector3,
    triangle_normal: math.Vector3) -> Tuple[math.Vector3, math.Vector3]:
  """Gets the closest points between a line segment and triangle."""
  # The closest triangle point is either on the edge or within the triangle.
  # First check triangle edges for the closest point.
  seg_pt1, tri_pt1 = closest_segment_to_segment_points(a, b, p0, p1)
  d1 = (seg_pt1 - tri_pt1).dot(seg_pt1 - tri_pt1)
  seg_pt2, tri_pt2 = closest_segment_to_segment_points(a, b, p1, p2)
  d2 = (seg_pt2 - tri_pt2).dot(seg_pt2 - tri_pt2)
  seg_pt3, tri_pt3 = closest_segment_to_segment_points(a, b, p0, p2)
  d3 = (seg_pt3 - tri_pt3).dot(seg_pt3 - tri_pt3)

  # Next, handle the case where the closest triangle point is inside the
  # triangle. Either the line segment intersects the triangle or a segment
  # endpoint is closest to a point inside the triangle.
  # If the line overlaps the triangle and is parallel to the triangle plane,
  # the chosen triangle point is arbitrary.
  seg_pt4 = closest_segment_point_plane(a, b, p0, triangle_normal)
  tri_pt4 = closest_triangle_point(p0, p1, p2, seg_pt4)
  d4 = (seg_pt4 - tri_pt4).dot(seg_pt4 - tri_pt4)

  # Get the point with minimum distance from the line segment point to the
  # triangle point.
  distance = jp.array([[d1, d2, d3, d4]])
  min_dist = jp.amin(distance)
  mask = (distance == min_dist).T
  seg_pt = jp.array([seg_pt1, seg_pt2, seg_pt3, seg_pt4]) * mask
  tri_pt = jp.array([tri_pt1, tri_pt2, tri_pt3, tri_pt4]) * mask
  seg_pt = jp.sum(seg_pt, axis=0) / jp.sum(mask)
  tri_pt = jp.sum(tri_pt, axis=0) / jp.sum(mask)
  return seg_pt, tri_pt


def _project_pt_onto_plane(pt: math.Vector3, plane_pt: math.Vector3,
                           plane_normal: math.Vector3) -> math.Vector3:
  """Projects a point onto a plane."""
  dist = (pt - plane_pt).dot(plane_normal)
  return pt - dist * plane_normal


def _project_poly_onto_plane(poly: jp.ndarray, plane_pt: math.Vector3,
                             plane_normal: math.Vector3) -> jp.ndarray:
  """Projects a polygon onto a plane."""
  return jp.vmap(
      _project_pt_onto_plane, include=[True, False,
                                       False])(poly, plane_pt,
                                               math.normalize(plane_normal))


def _project_poly_onto_poly_plane(poly1: math.Vector3, norm1: math.Vector3,
                                  poly2: math.Vector3,
                                  norm2: math.Vector3) -> math.Vector3:
  """"Projects poly1 onto the poly2 plane along norm1."""
  d = poly2[0].dot(norm2)
  denom = norm1.dot(norm2)
  t = (d - poly1.dot(norm2)) / (denom + 1e-6 * (denom == 0.))
  new_poly = poly1 + t.reshape(-1, 1) * norm1
  return new_poly


def point_in_front_of_plane(plane_pt: math.Vector3, plane_normal: math.Vector3,
                            pt: math.Vector3) -> bool:
  return (pt - plane_pt).dot(plane_normal) > 1e-6


def get_orthogonals(a):
  """Given a normal vector `a`, returns orthogonal vectors `b` and `c`."""
  # TODO reconcile `get_orthogonals` with the impl in `math.py`.
  a_abs = jp.abs(a)
  b = jp.ones_like(a)
  idx = jp.argmax(a_abs)
  denom = a[idx] + 1e-6 * (a[idx] == 0.)
  b = jp.index_update(b, idx, -(a.sum(axis=0) - a[idx]) / denom)
  c = jp.cross(a, b)
  return b, c


def _clip_edge_to_planes(edge_p0, edge_p1, plane_pts, plane_normals):
  """Clips an edge to planes."""
  # We return two clipped points, and a mask to include the new edge or not.
  p0, p1 = edge_p0, edge_p1
  p0_in_front = jp.vmap(jp.dot)(p0 - plane_pts, plane_normals) > 1e-6
  p1_in_front = jp.vmap(jp.dot)(p1 - plane_pts, plane_normals) > 1e-6

  # Get candidate clipped points along line segment (p0, p1) by clipping against
  # all clipping planes.
  candidate_clipped_ps = jp.vmap(
      closest_segment_point_plane,
      include=[False, False, True, True])(p0, p1, plane_pts, plane_normals)

  def clip_edge_point(p0, p1, p0_in_front, clipped_ps):

    @jp.vmap
    def choose_edge_point(in_front, clipped_p):
      return jp.where(in_front, clipped_p, p0)

    # Pick the clipped point if p0 is in front of the clipping plane. Otherwise
    # keep p0 as the edge point.
    new_edge_ps = choose_edge_point(p0_in_front, clipped_ps)

    # Pick the clipped point that is most along the edge direction.
    # This degenerates to picking the original point p0 if p0 is *not* in front
    # of any clipping planes.
    dists = jp.dot(new_edge_ps - p0, p1 - p0)
    new_edge_p = new_edge_ps[jp.argmax(dists)]
    return new_edge_p

  # Clip each edge point.
  new_p0 = clip_edge_point(p0, p1, p0_in_front, candidate_clipped_ps)
  new_p1 = clip_edge_point(p1, p0, p1_in_front, candidate_clipped_ps)
  clipped_pts = jp.array([new_p0, new_p1])

  # Keep the original points if both points are in front of any of the clipping
  # planes, rather than creating a new clipped edge. If the entire subject edge
  # is in front of any clipping plane, we need to grab an edge from the clipping
  # polygon instead.
  both_in_front = p0_in_front & p1_in_front
  mask = ~jp.any(both_in_front)
  new_ps = jp.where(mask, clipped_pts, jp.array([p0, p1]))
  # Mask out crossing clipped edge points.
  mask = jp.where((p0 - p1).dot(new_ps[0] - new_ps[1]) < 0, False, mask)
  return new_ps, jp.array([mask, mask])


def clip(clipping_poly: math.Vector3, subject_poly: math.Vector3,
         clipping_normal: math.Vector3,
         subject_normal: math.Vector3) -> Tuple[math.Vector3, jp.ndarray]:
  """Clips a subject polygon against a clipping polygon.

  A parallelized clipping algorithm for convex polygons. The result is a set of
  vertices on the clipped subject polygon in the subject polygon plane.

  Args:
    clipping_poly: The polygon that we use to clip the subject polygon against.
    subject_poly: The polygon that gets clipped.
    clipping_normal: Normal of the clipping polygon.
    subject_normal: Normal of the subject polygon.

  Returns:
    clipped_pts: The clipped polygon.
    mask: True if a point is in the clipping polygon.
  """
  # Get clipping edge points, edge planes, and edge normals.
  clipping_p0 = jp.roll(clipping_poly, 1, axis=0)
  clipping_plane_pts = clipping_p0
  clipping_p1 = clipping_poly
  clipping_plane_normals = jp.vmap(
      jp.cross, include=[False, True])(clipping_normal,
                                       clipping_p1 - clipping_p0)

  # Get subject edge points, edge planes, and edge normals.
  subject_edge_p0 = jp.roll(subject_poly, 1, axis=0)
  subject_plane_pts = subject_edge_p0
  subject_edge_p1 = subject_poly
  subject_plane_normals = jp.vmap(
      jp.cross, include=[False, True])(subject_normal,
                                       subject_edge_p1 - subject_edge_p0)

  # Clip all edges of the subject poly against clipping planes.
  clipped_edges0, masks0 = jp.vmap(
      _clip_edge_to_planes,
      include=[True, True, False,
               False])(subject_edge_p0, subject_edge_p1, clipping_plane_pts,
                       clipping_plane_normals)

  # Project the clipping poly onto the subject plane.
  clipping_p0_s = _project_poly_onto_poly_plane(clipping_p0, clipping_normal,
                                                subject_poly, subject_normal)
  clipping_p1_s = _project_poly_onto_poly_plane(clipping_p1, clipping_normal,
                                                subject_poly, subject_normal)
  # Clip all edges of the clipping poly against subject planes.
  clipped_edges1, masks1 = jp.vmap(
      _clip_edge_to_planes,
      include=[True, True, False,
               False])(clipping_p0_s, clipping_p1_s, subject_plane_pts,
                       subject_plane_normals)

  # Merge the points and reshape.
  clipped_edges = jp.concatenate([clipped_edges0, clipped_edges1])
  masks = jp.concatenate([masks0, masks1])
  clipped_points = clipped_edges.reshape((-1, 3))
  mask = masks.reshape(-1)

  return clipped_points, mask


def _create_sat_edge_contact(a0: math.Vector3, a1: math.Vector3,
                             b0: math.Vector3, b1: math.Vector3,
                             maybe_edge_contact: bool, sep_axis: math.Vector3,
                             sep_dist: float) -> Contact:
  """Creates an edge contact from a separating axis."""
  # We get the closest segment-segment points between the two edges that are
  # creating the separating axis.
  an, bn, ta, tb = _closest_segment_to_segment_points(a0, a1, b0, b1)
  valid_edge_contact = (
      maybe_edge_contact & (ta >= 0) & (ta <= 1) & (tb >= 0) & (tb <= 1))
  pos = jp.where(valid_edge_contact, bn + (an - bn) * 0.5, jp.zeros_like(a0))
  normal = -sep_axis
  penetration = jp.where(valid_edge_contact, -sep_dist, -jp.ones_like(sep_dist))
  contact = Contact(
      pos=pos, vel=jp.zeros_like(pos), normal=normal, penetration=penetration)
  # Create 4 contact points, even though we only use the first one. The first
  # contact point is from the edge contact, and the other three are padding
  # since SAT creates up to four contact points.
  mask = jp.array([valid_edge_contact, False, False, False])
  contact = jp.tree_map(lambda x: jp.stack([x] * 4, axis=0), contact)
  contact.penetration = jp.where(mask, penetration, -jp.ones_like(penetration))
  return contact


def _create_sat_contact_manifold(clipping_poly: jp.ndarray,
                                 subject_poly: jp.ndarray,
                                 clipping_norm: math.Vector3,
                                 subject_norm: math.Vector3,
                                 sep_axis_sign: int) -> Contact:
  """Creates a contact manifold between two convex polygons.

  The polygon faces are expected to have a clockwise winding order so that
  clipping plane normals point away from the polygon center.

  Args:
    clipping_poly: The reference polygon to clip the contact against.
    subject_poly: The subject polygon to clip contacts onto.
    clipping_norm: The clipping polygon normal.
    subject_norm: The subject polygon normal.
    sep_axis_sign: The sign of the separating axis wrt the clipping polygon
      normal.

  Returns:
    contact: Contact object.
  """
  # Clip the subject (incident) face onto the clipping (reference) face.
  # The incident points are clipped points on the subject polygon.
  poly_incident, mask = clip(clipping_poly, subject_poly, clipping_norm,
                             subject_norm)
  # The reference points are clipped points on the clipping polygon.
  poly_ref = _project_poly_onto_plane(poly_incident, clipping_poly[0],
                                      clipping_norm)
  behind_clipping_plane = point_in_front_of_plane(clipping_poly[0],
                                                  -clipping_norm, poly_incident)
  mask = mask & behind_clipping_plane

  # TODO: consider changing this to maximize the manifold area.
  ortho_1, ortho_2 = get_orthogonals(clipping_norm)
  dist_mask = jp.where(mask, 0.0, -1e6)
  best_x = jp.argmax(poly_ref.dot(ortho_1) + dist_mask)
  best_nx = jp.argmax(poly_ref.dot(-ortho_1) + dist_mask)
  best_y = jp.argmax(poly_ref.dot(ortho_2) + dist_mask)
  best_ny = jp.argmax(poly_ref.dot(-ortho_2) + dist_mask)
  # Choose up to four contact points.
  best = jp.array([best_x, best_nx, best_y, best_ny])
  contact_pts = jp.take(poly_ref, best)
  mask_pts = jp.take(mask, best)
  penetration_dir = jp.take(poly_incident, best) - contact_pts
  penetration = penetration_dir.dot(-clipping_norm)
  penetration = jp.where(mask_pts, penetration, -jp.ones_like(penetration))

  contact = Contact(
      pos=contact_pts,
      vel=-jp.ones_like(contact_pts),
      normal=jp.stack([sep_axis_sign * clipping_norm] * 4, 0),
      penetration=penetration)

  return contact


def sat_hull_hull(faces_a: jp.ndarray, faces_b: jp.ndarray,
                  vertices_a: jp.ndarray, vertices_b: jp.ndarray,
                  normals_a: jp.ndarray,
                  normals_b: jp.ndarray) -> Tuple[Contact, Contact]:
  """Runs the Separating Axis Test for a pair of hulls.

  Given two convex hulls, the Separating Axis Test finds a separating axis
  between all edge pairs and face pairs. Edge pairs create a single contact
  point and face pairs create a contact manifold (up to four contact points).
  We return both the edge and face contacts. Valid contacts can be checked with
  penetration > 0. Resulting edge contacts should be preferred over face
  contacts.

  Args:
    faces_a: An ndarray of hull A's polygon faces.
    faces_b: An ndarray of hull B's polygon faces.
    vertices_a: Vertices for hull A.
    vertices_b: Vertices for hull B.
    normals_a: Normal vectors for hull A's polygon faces.
    normals_b: Normal vectors for hull B's polygon faces.

  Returns:
    edge_contact: An edge contact.
    face_contact: A face contact.
  """

  @jp.vmap
  def get_edge_axes(poly1, poly2):
    # TODO: consider caching/de-duping edges at system creation time.
    points11 = poly1
    points12 = jp.roll(poly1, 1, axis=0)
    poly_edges1 = points11 - points12
    points21 = poly2
    points22 = jp.roll(poly2, 1, axis=0)
    poly_edges2 = points21 - points22

    edges1 = jp.tile(poly_edges1, reps=(poly_edges2.shape[0], 1))
    points11p = jp.tile(points11, reps=(poly_edges2.shape[0], 1))
    points12p = jp.tile(points12, reps=(poly_edges2.shape[0], 1))

    edges2 = jp.repeat(poly_edges2, repeats=poly_edges1.shape[0], axis=0)
    points21p = jp.repeat(points21, repeats=poly_edges1.shape[0], axis=0)
    points22p = jp.repeat(points22, repeats=poly_edges1.shape[0], axis=0)

    edge_edge_axes = jp.vmap(jp.cross)(edges1, edges2)
    directions = jp.vmap(jp.dot)(points11p - origin_a, edge_edge_axes)
    directions = directions.reshape(-1, 1)
    edge_edge_axes *= jp.where(directions > 0, 1, -1)
    return edge_edge_axes, points11p, points12p, points21p, points22p

  def get_face_support(vertices_b, plane_normals_a, plane_points_a):
    dists = jp.vmap(
        lambda n, p, s: jp.vmap(jp.dot)(n, s - p),
        include=[False, False, True])(plane_normals_a, plane_points_a,
                                      vertices_b)
    dists = jp.amin(dists, axis=0)
    best_dist_idx = jp.argmax(dists, axis=0)
    return dists[best_dist_idx], best_dist_idx

  def get_edge_support(vertices_b, plane_normals_a, plane_points_a, edge_b1,
                       edge_b2, aux_dists, mask):
    dists = jp.vmap(
        lambda n, p, s: jp.vmap(jp.dot)(n, s - p),
        include=[False, False, True])(plane_normals_a, plane_points_a,
                                      vertices_b)
    support_idx = jp.argmin(dists, axis=0)
    support_point = vertices_b[support_idx]
    support_dist = jp.vmap(jp.take)(dists.T, support_idx)

    # Mask support points if they are not on the edge that created the
    # separating axis.
    support_point_on_edge_b1 = (edge_b1 - support_point).sum(axis=1) == 0
    support_point_on_edge_b2 = (edge_b2 - support_point).sum(axis=1) == 0
    mask = mask | ~(support_point_on_edge_b1 | support_point_on_edge_b2)

    support_dist = jp.where(mask, -1e6, support_dist)
    best_dist_idx = jp.argmax(support_dist + aux_dists, axis=0)
    return support_dist[best_dist_idx], best_dist_idx

  # TODO: considering making the origin and edges static.
  origin_a = jp.mean(vertices_a, axis=0)

  # Get all face normal candidate sep axes, then find the face with min
  # separation.
  dist1, idx1 = get_face_support(vertices_a, normals_b, faces_b[:, 0, :])
  dist2, idx2 = get_face_support(vertices_b, normals_a, faces_a[:, 0, :])
  face_dist = jp.where(dist1 > dist2, dist1, dist2)
  face_idx = jp.where(dist1 > dist2, idx1, idx2)
  ref_face = jp.where(dist1 > dist2, faces_b[face_idx], faces_a[face_idx])
  ref_face_norm = jp.where(dist1 > dist2, normals_b[face_idx],
                           normals_a[face_idx])
  sep_axis_sign = jp.where(dist1 > dist2, 1, -1)

  # Get the incident (most antiparallel) face on the other hull.
  incident_faces = jp.where(dist1 > dist2, faces_a, faces_b)
  incident_face_norms = jp.where(dist1 > dist2, normals_a, normals_b)
  dirs = jp.vmap(
      jp.dot, include=[True, False])(incident_face_norms, ref_face_norm)
  incident_face_idx = jp.argmin(dirs)
  incident_face = incident_faces[incident_face_idx]
  incident_face_norm = incident_face_norms[incident_face_idx]

  # Create potential face contact.
  face_contact = _create_sat_contact_manifold(ref_face, incident_face,
                                              ref_face_norm, incident_face_norm,
                                              sep_axis_sign)

  # Get all edge candidate separating axes by enumerating all edge pairs.
  num_faces_a, num_faces_b = faces_a.shape[0], faces_b.shape[0]
  faces_a_r = jp.tile(faces_a, reps=(num_faces_a, 1, 1))
  faces_b_r = jp.repeat(faces_b, repeats=num_faces_b, axis=0)
  edge_normals, edge_a1, edge_a2, edge_b1, edge_b2 = jp.tree_map(
      jp.concatenate, get_edge_axes(faces_a_r, faces_b_r))

  # Get rid of separating axes created by paralell edges.
  bad_edge_axes = jp.all(edge_normals == 0., axis=1)
  # Get rid of separating axes that don't create valid separating planes.
  self_dists = jp.vmap(
      lambda n, p, s: jp.vmap(jp.dot)(n, s - p),
      include=[False, False, True])(edge_normals, edge_a1, vertices_a)
  sep_plane_intersects_hull_a = jp.amax(self_dists, axis=0) > 0
  bad_edge_axes = bad_edge_axes | sep_plane_intersects_hull_a
  # Get the distance between edge-pairs to use in the get_edge_support function.
  edges_a_mid = edge_a1 + (edge_a2 - edge_a1) * 0.5
  edges_b_mid = edge_b1 + (edge_b2 - edge_b1) * 0.5
  midpoint_dists = -jp.vmap(jp.dot)(edges_a_mid - edges_b_mid,
                                    edges_a_mid - edges_b_mid)

  # Find the edge-pair with min separation.
  edge_dist, edge_idx = get_edge_support(vertices_b, edge_normals, edge_a1,
                                         edge_b1, edge_b2, midpoint_dists,
                                         bad_edge_axes)
  edge_normal = edge_normals[edge_idx]
  maybe_edge_contact = edge_dist > face_dist
  edge_normal = edge_normal / jp.safe_norm(edge_normal)

  best_dist = jp.amax(jp.array([edge_dist, face_dist]))
  has_intersection = best_dist < 0

  # Create potential edge contact.
  edge_contact = _create_sat_edge_contact(edge_a1[edge_idx], edge_a2[edge_idx],
                                          edge_b1[edge_idx], edge_b2[edge_idx],
                                          has_intersection & maybe_edge_contact,
                                          edge_normal, edge_dist)

  return edge_contact, face_contact


def rotate_point_axis(qp: QP, rotate_idxs: jp.ndarray, point: jp.ndarray,
                      axis: jp.ndarray, angle: float):
  """Convenience function for rotating qps in world space around a point.

  For a selection of body indices, this function takes those bodies and
  rotates them around some point, around an axis emanating from that point.

  Args:
    qp: (n, ...) Full state qp for system
    rotate_idxs: (m<=n,) integer body idxs to be rotated
    point: (3,) a point in space to rotate about
    axis: (3,) an axis in space to rotate around
    angle: float angle to rotate by

  Returns:
    Full state qp with bodies at rotate_idxs rotated.
  """
  qrot = jp.array(qp.rot.shape[0] * [[1., 0., 0., 0.]])
  qrot[rotate_idxs] = math.quat_rot_axis(axis, angle)
  disp = qp.pos - point
  rotated_pos = jp.vmap(math.rotate)(disp, qrot) + point
  rotated_rot = jp.vmap(math.quat_mul)(qrot, qp.rot)
  qp = qp.replace(pos=rotated_pos, rot=rotated_rot)
  return qp



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Augmented Random Search training tests."""
import pickle

from absl.testing import absltest
from absl.testing import parameterized
from brax import envs
from brax.training.acme import running_statistics
from brax.training.agents.ars import networks as ars_networks
from brax.training.agents.ars import train as ars
import jax


class ARSTest(parameterized.TestCase):
  """Tests for ARS module."""

  @parameterized.parameters(True, False)
  def testModelEncoding(self, normalize_observations):
    env = envs.get_environment('fast')
    _, params, _ = ars.train(
        env,
        num_timesteps=128,
        episode_length=128,
        normalize_observations=normalize_observations)
    normalize_fn = lambda x, y: x
    if normalize_observations:
      normalize_fn = running_statistics.normalize
    ars_network = ars_networks.make_policy_network(env.observation_size,
                                                   env.action_size,
                                                   normalize_fn)
    inference = ars_networks.make_inference_fn(ars_network)
    byte_encoding = pickle.dumps(params)
    decoded_params = pickle.loads(byte_encoding)

    # Compute one action.
    state = env.reset(jax.random.PRNGKey(0))
    action = inference(decoded_params)(state.obs, jax.random.PRNGKey(0))[0]
    env.step(state, action)


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for geometry."""

from absl.testing import absltest
from brax.v2 import math
import jax
from jax import numpy as jp
import numpy as np


class MathTest(absltest.TestCase):

  def test_inv_approximate(self):
    # create a 4x4 matrix we know is invertible
    x = jax.random.normal(jax.random.PRNGKey(0), (4, 4))
    x = jp.eye(4) * 0.001 + x @ x.T

    x_inv = jp.linalg.inv(x)
    x_inv_approximate = math.inv_approximate(x, jp.zeros((4, 4)), maxiter=100)

    np.testing.assert_array_almost_equal(x_inv_approximate, x_inv)

  def test_from_to(self):
    v1 = jp.array([1.0, 0.0, 0.0])
    rot = math.from_to(v1, v1)
    np.testing.assert_array_almost_equal(rot, jp.array([1.0, 0.0, 0.0, 0.0]))

    rot = math.from_to(v1, -v1)
    np.testing.assert_array_almost_equal(rot, jp.array([0.0, 0.0, 0.0, 1.0]))

    rot = math.from_to(-v1, v1)
    np.testing.assert_array_almost_equal(rot, jp.array([0.0, 0.0, 0.0, -1.0]))

    v2 = jp.array([0.0, 1.0, 0.0])
    rot = math.from_to(v2, -v2)
    np.testing.assert_array_almost_equal(rot, jp.array([0.0, 0.0, 0.0, -1.0]))

    v2 = jp.array([-0.5, 0.5, 0.0])
    v2 /= jp.linalg.norm(v2)
    rot = math.from_to(v1, v2)
    np.testing.assert_array_almost_equal(
        rot, math.euler_to_quat(jp.array([0.0, 0.0, 135.0]))
    )


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Proximal policy optimization training.

See: https://arxiv.org/pdf/1707.06347.pdf
"""

import functools
import time
from typing import Callable, Optional, Tuple

from absl import logging
from brax import envs
from brax.envs import wrappers
from brax.training import acting
from brax.training import gradients
from brax.training import pmap
from brax.training import types
from brax.training.acme import running_statistics
from brax.training.acme import specs
from brax.training.agents.ppo import losses as ppo_losses
from brax.training.agents.ppo import networks as ppo_networks
from brax.training.types import Params
from brax.training.types import PRNGKey
import flax
import jax
import jax.numpy as jnp
import optax

InferenceParams = Tuple[running_statistics.NestedMeanStd, Params]
Metrics = types.Metrics

_PMAP_AXIS_NAME = 'i'


@flax.struct.dataclass
class TrainingState:
  """Contains training state for the learner."""
  optimizer_state: optax.OptState
  params: ppo_losses.PPONetworkParams
  normalizer_params: running_statistics.RunningStatisticsState
  env_steps: jnp.ndarray


def _unpmap(v):
  return jax.tree_util.tree_map(lambda x: x[0], v)


def train(environment: envs.Env,
          num_timesteps: int,
          episode_length: int,
          action_repeat: int = 1,
          num_envs: int = 1,
          max_devices_per_host: Optional[int] = None,
          num_eval_envs: int = 128,
          learning_rate: float = 1e-4,
          entropy_cost: float = 1e-4,
          discounting: float = 0.9,
          seed: int = 0,
          unroll_length: int = 10,
          batch_size: int = 32,
          num_minibatches: int = 16,
          num_updates_per_batch: int = 2,
          num_evals: int = 1,
          normalize_observations: bool = False,
          reward_scaling: float = 1.,
          clipping_epsilon: float = .3,
          gae_lambda: float = .95,
          deterministic_eval: bool = False,
          network_factory: types.NetworkFactory[
              ppo_networks.PPONetworks] = ppo_networks.make_ppo_networks,
          progress_fn: Callable[[int, Metrics], None] = lambda *args: None,
          normalize_advantage: bool = True,
          eval_env: Optional[envs.Env] = None):
  """PPO training."""
  assert batch_size * num_minibatches % num_envs == 0
  xt = time.time()

  process_count = jax.process_count()
  process_id = jax.process_index()
  local_device_count = jax.local_device_count()
  local_devices_to_use = local_device_count
  if max_devices_per_host:
    local_devices_to_use = min(local_devices_to_use, max_devices_per_host)
  logging.info(
      'Device count: %d, process count: %d (id %d), local device count: %d, '
      'devices to be used count: %d', jax.device_count(), process_count,
      process_id, local_device_count, local_devices_to_use)
  device_count = local_devices_to_use * process_count

  # The number of environment steps executed for every training step.
  env_step_per_training_step = (
      batch_size * unroll_length * num_minibatches * action_repeat)
  num_evals_after_init = max(num_evals - 1, 1)
  # The number of training_step calls per training_epoch call.
  # equals to ceil(num_timesteps / (num_evals * env_step_per_training_step))
  num_training_steps_per_epoch = -(
      -num_timesteps // (num_evals_after_init * env_step_per_training_step))

  assert num_envs % device_count == 0
  env = environment

  env = wrappers.wrap_for_training(
      env, episode_length=episode_length, action_repeat=action_repeat)

  reset_fn = jax.jit(jax.vmap(env.reset))

  normalize = lambda x, y: x
  if normalize_observations:
    normalize = running_statistics.normalize
  ppo_network = network_factory(
      env.observation_size,
      env.action_size,
      preprocess_observations_fn=normalize)
  make_policy = ppo_networks.make_inference_fn(ppo_network)

  optimizer = optax.adam(learning_rate=learning_rate)

  loss_fn = functools.partial(
      ppo_losses.compute_ppo_loss,
      ppo_network=ppo_network,
      entropy_cost=entropy_cost,
      discounting=discounting,
      reward_scaling=reward_scaling,
      gae_lambda=gae_lambda,
      clipping_epsilon=clipping_epsilon,
      normalize_advantage=normalize_advantage)

  gradient_update_fn = gradients.gradient_update_fn(
      loss_fn, optimizer, pmap_axis_name=_PMAP_AXIS_NAME, has_aux=True)

  def minibatch_step(
      carry, data: types.Transition,
      normalizer_params: running_statistics.RunningStatisticsState):
    optimizer_state, params, key = carry
    key, key_loss = jax.random.split(key)
    (_, metrics), params, optimizer_state = gradient_update_fn(
        params,
        normalizer_params,
        data,
        key_loss,
        optimizer_state=optimizer_state)

    return (optimizer_state, params, key), metrics

  def sgd_step(carry, unused_t, data: types.Transition,
               normalizer_params: running_statistics.RunningStatisticsState):
    optimizer_state, params, key = carry
    key, key_perm, key_grad = jax.random.split(key, 3)

    def convert_data(x: jnp.ndarray):
      x = jax.random.permutation(key_perm, x)
      x = jnp.reshape(x, (num_minibatches, -1) + x.shape[1:])
      return x

    shuffled_data = jax.tree_util.tree_map(convert_data, data)
    (optimizer_state, params, _), metrics = jax.lax.scan(
        functools.partial(minibatch_step, normalizer_params=normalizer_params),
        (optimizer_state, params, key_grad),
        shuffled_data,
        length=num_minibatches)
    return (optimizer_state, params, key), metrics

  def training_step(
      carry: Tuple[TrainingState, envs.State, PRNGKey],
      unused_t) -> Tuple[Tuple[TrainingState, envs.State, PRNGKey], Metrics]:
    training_state, state, key = carry
    key_sgd, key_generate_unroll, new_key = jax.random.split(key, 3)

    policy = make_policy(
        (training_state.normalizer_params, training_state.params.policy))

    def f(carry, unused_t):
      current_state, current_key = carry
      current_key, next_key = jax.random.split(current_key)
      next_state, data = acting.generate_unroll(
          env,
          current_state,
          policy,
          current_key,
          unroll_length,
          extra_fields=('truncation',))
      return (next_state, next_key), data

    (state, _), data = jax.lax.scan(
        f, (state, key_generate_unroll), (),
        length=batch_size * num_minibatches // num_envs)
    # Have leading dimentions (batch_size * num_minibatches, unroll_length)
    data = jax.tree_util.tree_map(lambda x: jnp.swapaxes(x, 1, 2), data)
    data = jax.tree_util.tree_map(lambda x: jnp.reshape(x, (-1,) + x.shape[2:]),
                                  data)
    assert data.discount.shape[1:] == (unroll_length,)

    # Update normalization params and normalize observations.
    normalizer_params = running_statistics.update(
        training_state.normalizer_params,
        data.observation,
        pmap_axis_name=_PMAP_AXIS_NAME)

    (optimizer_state, params, _), metrics = jax.lax.scan(
        functools.partial(
            sgd_step, data=data, normalizer_params=normalizer_params),
        (training_state.optimizer_state, training_state.params, key_sgd), (),
        length=num_updates_per_batch)

    new_training_state = TrainingState(
        optimizer_state=optimizer_state,
        params=params,
        normalizer_params=normalizer_params,
        env_steps=training_state.env_steps + env_step_per_training_step)
    return (new_training_state, state, new_key), metrics

  def training_epoch(training_state: TrainingState, state: envs.State,
                     key: PRNGKey) -> Tuple[TrainingState, envs.State, Metrics]:
    (training_state, state, _), loss_metrics = jax.lax.scan(
        training_step, (training_state, state, key), (),
        length=num_training_steps_per_epoch)
    loss_metrics = jax.tree_util.tree_map(jnp.mean, loss_metrics)
    return training_state, state, loss_metrics

  training_epoch = jax.pmap(training_epoch, axis_name=_PMAP_AXIS_NAME)

  # Note that this is NOT a pure jittable method.
  def training_epoch_with_timing(
      training_state: TrainingState, env_state: envs.State,
      key: PRNGKey) -> Tuple[TrainingState, envs.State, Metrics]:
    nonlocal training_walltime
    t = time.time()
    (training_state, env_state,
     metrics) = training_epoch(training_state, env_state, key)
    metrics = jax.tree_util.tree_map(jnp.mean, metrics)
    jax.tree_util.tree_map(lambda x: x.block_until_ready(), metrics)

    epoch_training_time = time.time() - t
    training_walltime += epoch_training_time
    sps = (num_training_steps_per_epoch *
           env_step_per_training_step) / epoch_training_time
    metrics = {
        'training/sps': sps,
        'training/walltime': training_walltime,
        **{f'training/{name}': value for name, value in metrics.items()}
    }
    return training_state, env_state, metrics

  key = jax.random.PRNGKey(seed)
  global_key, local_key = jax.random.split(key)
  del key
  local_key = jax.random.fold_in(local_key, process_id)
  local_key, key_env, eval_key = jax.random.split(local_key, 3)
  # key_networks should be global, so that networks are initialized the same
  # way for different processes.
  key_policy, key_value = jax.random.split(global_key)
  del global_key

  init_params = ppo_losses.PPONetworkParams(
      policy=ppo_network.policy_network.init(key_policy),
      value=ppo_network.value_network.init(key_value))
  training_state = TrainingState(
      optimizer_state=optimizer.init(init_params),
      params=init_params,
      normalizer_params=running_statistics.init_state(
          specs.Array((env.observation_size,), jnp.float32)),
      env_steps=0)
  training_state = jax.device_put_replicated(
      training_state,
      jax.local_devices()[:local_devices_to_use])

  key_envs = jax.random.split(key_env, num_envs // process_count)
  key_envs = jnp.reshape(key_envs,
                         (local_devices_to_use, -1) + key_envs.shape[1:])
  env_state = reset_fn(key_envs)

  if not eval_env:
    eval_env = env
  else:
    eval_env = wrappers.wrap_for_training(
        eval_env, episode_length=episode_length, action_repeat=action_repeat)

  evaluator = acting.Evaluator(
      eval_env,
      functools.partial(make_policy, deterministic=deterministic_eval),
      num_eval_envs=num_eval_envs,
      episode_length=episode_length,
      action_repeat=action_repeat,
      key=eval_key)

  # Run initial eval
  if process_id == 0 and num_evals > 1:
    metrics = evaluator.run_evaluation(
        _unpmap(
            (training_state.normalizer_params, training_state.params.policy)),
        training_metrics={})
    logging.info(metrics)
    progress_fn(0, metrics)

  training_walltime = 0
  current_step = 0
  for it in range(num_evals_after_init):
    logging.info('starting iteration %s %s', it, time.time() - xt)

    # optimization
    epoch_key, local_key = jax.random.split(local_key)
    epoch_keys = jax.random.split(epoch_key, local_devices_to_use)
    (training_state, env_state,
     training_metrics) = training_epoch_with_timing(training_state, env_state,
                                                    epoch_keys)
    current_step = int(_unpmap(training_state.env_steps))

    if process_id == 0:
      # Run evals.
      metrics = evaluator.run_evaluation(
          _unpmap(
              (training_state.normalizer_params, training_state.params.policy)),
          training_metrics)
      logging.info(metrics)
      progress_fn(current_step, metrics)

  total_steps = current_step
  assert total_steps >= num_timesteps

  # If there was no mistakes the training_state should still be identical on all
  # devices.
  pmap.assert_is_replicated(training_state)
  params = _unpmap(
      (training_state.normalizer_params, training_state.params.policy))
  logging.info('total steps: %s', total_steps)
  pmap.synchronize_hosts()
  return (make_policy, params, metrics)



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Functions for forward and inverse kinematics."""
from typing import Tuple

from brax.v2 import base
from brax.v2 import math
from brax.v2 import scan
from brax.v2.base import Motion, System, Transform
import jax
from jax import numpy as jp


def forward(
    sys: System, q: jp.ndarray, qd: jp.ndarray
) -> Tuple[Transform, Motion]:
  """Converts joint position/velocity to transform/motion in world frame.

  Args:
    sys: system defining the kinematic tree and other properties
    q: joint position vector
    qd: joint velocity vector

  Returns:
    x: transform in world frame
    xd: motion in world frame
  """

  # convert joint position/velocity to transform/motion in joint frame
  def jcalc(typ, q, qd, motion):
    if typ == 'f':
      q, qd = q.reshape((-1, 7)), qd.reshape((-1, 6))
      j = Transform(pos=q[:, 0:3], rot=q[:, 3:7])
      jd = Motion(ang=qd[:, 3:6], vel=qd[:, 0:3])
    else:
      # create joint transforms and motions:
      # - rotation/velocity about axis for revolute (motion.ang)
      # - translation/velocity along axis for prismatic (motion.vel)
      rot_fn = lambda ang, q: math.normalize(math.quat_rot_axis(ang, q))[0]
      j = Transform.create(
          rot=jax.vmap(rot_fn)(motion.ang, q),
          pos=jax.vmap(jp.multiply)(motion.vel, q),
      )
      jd = jax.vmap(lambda a, b: a * b)(motion, qd)

      # then group them by link, so each link has num_dofs joints
      num_links, num_dofs = qd.shape[0] // int(typ), int(typ)
      s = (num_links, num_dofs, -1)
      j_stack, jd_stack = j.reshape(s), jd.reshape(s)

      # accumulate j and jd one dof at a time
      j, jd = j_stack.take(0, axis=1), jd_stack.take(0, axis=1)
      for i in range(1, num_dofs):
        j_i, jd_i = j_stack.take(i, axis=1), jd_stack.take(i, axis=1)
        j = j.vmap().do(j_i)
        jd = jd + Motion(
            ang=jax.vmap(math.rotate)(jd_i.ang, j_i.rot),
            vel=jax.vmap(math.rotate)(
                jd_i.vel + jax.vmap(jp.cross)(j_i.pos, jd_i.ang), j_i.rot
            ),
        )

    return j, jd

  j, jd = scan.link_types(sys, jcalc, 'qdd', 'l', q, qd, sys.dof.motion)

  anchor = Transform.create(rot=j.rot).vmap().do(sys.link.joint)
  j = j.replace(pos=j.pos + sys.link.joint.pos - anchor.pos)  # joint pos offset
  j = sys.link.transform.vmap().do(j)  # link transform

  def world(parent, j, jd):
    """Convert transform/motion from joint frame to world frame."""
    if parent is None:
      return j, jd
    x, xd = parent
    # TODO: determine why the motion `do` is inverted
    x = x.vmap().do(j)
    xd = xd + Motion(
        ang=jax.vmap(math.rotate)(jd.ang, x.rot),
        vel=jax.vmap(math.rotate)(
            jd.vel + jax.vmap(jp.cross)(x.pos, jd.ang), x.rot
        ),
    )
    return x, xd

  x, xd = scan.tree(sys, world, 'll', j, jd)

  x = x.replace(rot=jax.vmap(math.normalize)(x.rot)[0])

  return x, xd


def world_to_joint_frame(
    sys: System, x: Transform, xd: Motion
) -> Tuple[Transform, Motion]:
  """Moves into the joint frame of a maximal coordinates kinematic tree."""

  # pad a world transform onto the end
  x_pad = x.concatenate(Transform.zero((1,)))
  xd_pad = xd.concatenate(Motion.zero((1,)))

  p_idx = jp.array(sys.link_parents)

  x_p, xd_p = x_pad.take(p_idx), xd_pad.take(p_idx)
  x_c, xd_c = x, xd

  # move x and xd into joint frame
  x_joint = x_p.vmap().do(sys.link.transform).vmap().do(sys.link.joint)
  x_c = x_c.vmap().do(sys.link.joint)
  j = x_c.vmap().to_local(x_joint)

  # find world velocity of joint location point on parent
  xd_wj = Transform.create(pos=x_p.pos - x_joint.pos).vmap().do(xd_p)

  # move into joint coordinates
  xd_joint = xd_c - xd_wj
  inv_rotate = jax.vmap(lambda x, y: math.rotate(x, math.quat_inv(y)))
  jd = jax.tree_map(lambda x: inv_rotate(x, x_joint.rot), xd_joint)

  return j, jd


def link_to_joint_motion(motion: Motion) -> Tuple[Motion, float]:
  """Calculates 3-dof motions for joints corresponding to a given link motion.

  Args:
    motion: Motion dof with leaves of shape (QD_WIDTH, 3)

  Returns:
    Motion with all three dofs for this joint, and the handedness of the joint

  Joint axes are not always aligned along the local frame of a link. For
  example, a joint might be along the local z-axis, and we would need to
  find the rotation that takes local-x to local-z, or:
  from_to([1., 0., 0.], [0., 0., 1.]).

  For joints with more than 1 axis (e.g., universal, spherical), there's one
  additional degree of freedom in the orientation of the second axis, which is
  set by performing one final rotation of the link-frame y-axis so that it
  aligns with the joint-frame y-axis.

  Combined, these two rotations move the identity link-frame into the joint
  frame, where the new local-x and local-y point along the directions specified
  in the system dof.

  We also need translational components because the prismatic components of a
  joint might not be aligned with the rotational components of the joint.
  """

  # if this is already a 3-dof motion, then we're done, up to handedness
  if motion.ang.shape[0] == 3:
    return Motion(
        ang=jp.array([
            motion.ang[0],
            motion.ang[1],
            jp.cross(motion.ang[0], motion.ang[1]),
        ]),
        vel=jp.array([
            motion.vel[0],
            motion.vel[1],
            jp.cross(motion.vel[0], motion.vel[1]),
        ]),
    ), jp.dot(jp.cross(motion.ang[0], motion.ang[1]), motion.ang[2])

  # if motion is 1 or 2-dof, then we need to reconstruct the 3-dof frame that
  # completely defines its joint frame
  def j_axes(axis):
    frame = jp.eye(3)
    rot = math.from_to(frame[0], axis[0])
    y_temp = math.rotate(frame[1], rot)
    second_axis = axis[1] if axis.shape[0] > 1 else y_temp
    second_angle = math.signed_angle(axis[0], y_temp, second_axis)
    second_rot = math.quat_rot_axis(axis[0], second_angle)
    # if second axis is all zeros, then is an identity op
    rot = math.quat_mul(second_rot, rot)
    return rot

  link_rot_ang, link_rot_vel = j_axes(motion.ang), j_axes(motion.vel)

  def rotate_frame(axis_rotation):
    return jax.vmap(math.rotate, in_axes=(0, None))(jp.eye(3), axis_rotation)

  link_frame_ang = rotate_frame(link_rot_ang)
  link_frame_vel = rotate_frame(link_rot_vel)

  return Motion(ang=link_frame_ang, vel=link_frame_vel), 1.0


def axis_angle_ang(
    j: Transform, jd: Motion, motion: Motion
) -> Tuple[jp.ndarray, jp.ndarray, jp.ndarray, jp.ndarray]:
  """Returns axes, torque axes, angles, and angular velocities of a joint.

  Args:
    j: The transform for this joint.
    jd: The motion for this joint.
    motion: motion degrees of freedom for this joint

  Returns:
    Joint frame axis, torque axes, angles, and angular velocities.
  """
  joint_motion, parity = link_to_joint_motion(motion)

  v_rot = jax.vmap(math.rotate, in_axes=[0, None])
  child_frame = v_rot(joint_motion.ang, j.rot)

  line_of_nodes = jp.cross(child_frame[2], joint_motion.ang[0])
  line_of_nodes = line_of_nodes / (1e-10 + math.safe_norm(line_of_nodes))
  y_n_normal = joint_motion.ang[0]
  psi = math.signed_angle(y_n_normal, joint_motion.ang[1], line_of_nodes)
  axis_1_p_in_xz_c = (
      jp.dot(joint_motion.ang[0], child_frame[0]) * child_frame[0]
      + jp.dot(joint_motion.ang[0], child_frame[1]) * child_frame[1]
  )
  axis_1_p_in_xz_c = axis_1_p_in_xz_c / (
      1e-10 + math.safe_norm(axis_1_p_in_xz_c)
  )
  ang_between_1_p_xz_c = jp.dot(axis_1_p_in_xz_c, joint_motion.ang[0])
  theta = math.safe_arccos(jp.clip(ang_between_1_p_xz_c, -1, 1)) * jp.sign(
      jp.dot(joint_motion.ang[0], child_frame[2])
  )
  yc_n_normal = -child_frame[2] * parity
  phi = math.signed_angle(yc_n_normal, child_frame[1], line_of_nodes)

  axis = (child_frame[0], child_frame[1], child_frame[2] * parity)
  torque_axis = (joint_motion.ang[0], child_frame[1], child_frame[2] * parity)

  angle = (psi, theta, phi)
  vel = jax.tree_map(lambda x: jp.dot(x, jd.ang), axis)

  return axis, torque_axis, angle, vel


def axis_slide_vel(
    x: Transform, xd: Motion, motion: Motion
) -> Tuple[jp.ndarray, jp.ndarray, jp.ndarray]:
  """Returns axes and slide dofs for a joint.

  Args:
    x: (3,) The transform for this joint type.
    xd: (3,) The motion for this joint type.
    motion: Motion degrees of freedom for this joint

  Returns:
    Joint frame axis, positions, and velocities.
  """

  joint_motion, _ = link_to_joint_motion(motion)

  coords = joint_motion.vel @ x.pos
  velocities = joint_motion.vel @ xd.vel

  return joint_motion.vel, coords, velocities


def inverse(
    sys: System, x: Transform, xd: Motion
) -> Tuple[jp.ndarray, jp.ndarray]:
  """Translates maximal coordinates into reduced coordinates."""

  j, jd = world_to_joint_frame(sys, x, xd)

  def one_dof(j, jd, motion):
    _, _, (angle, _, _), (ang_vel, _, _) = axis_angle_ang(j, jd, motion)
    _, (slide_x, _, _), (vel, _, _) = axis_slide_vel(j, jd, motion)
    # TODO: investigate removing this `where``
    q = jp.where(motion.ang.any(), angle, slide_x)
    qd = jp.where(motion.ang.any(), ang_vel, vel)
    return q, qd

  def two_dof(j, jd, motion):
    _, _, angles, vels = axis_angle_ang(j, jd, motion)
    return jp.array(angles[0:2]), jp.array(vels[0:2])

  def three_dof(j, jd, motion):
    _, _, angles, vels = axis_angle_ang(j, jd, motion)
    return jp.array(angles[0:3]), jp.array(vels[0:3])

  def free(x, xd, _):
    return jp.concatenate([x.pos, x.rot]), jp.concatenate([xd.vel, xd.ang])

  def q_fn(typ, j, jd, motion):
    motion = jax.tree_map(
        lambda y: y.reshape((-1, base.QD_WIDTHS[typ], 3)), motion
    )
    q_fn_map = {
        'f': free,
        '1': one_dof,
        '2': two_dof,
        '3': three_dof,
    }

    q, qd = jax.vmap(q_fn_map[typ])(j, jd, motion)

    # transposed to preserve order of outputs
    return jp.array(q).reshape(-1), jp.array(qd).reshape(-1)

  q, qd = scan.link_types(sys, q_fn, 'lld', 'qd', j, jd, sys.dof.motion)
  return q, qd
